# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements realistic constraints to be used with SinglePeriodOptimization
and MultiPeriodOptimization policies, or other Cvxpy-based policies.
"""


import cvxpy as cp
import numpy as np

from .estimator import CvxpyExpressionEstimator, DataEstimator
from .forecast import HistoricalFactorizedCovariance

__all__ = [
    "LongOnly",
    "LeverageLimit",
    "LongCash",
    "DollarNeutral",
    "ParticipationRateLimit",
    "MaxWeights",
    "MinWeights",
    "FactorMaxLimit",
    "FactorMinLimit",
    "FixedFactorLoading",
    "MarketNeutral",
    "MinWeightsAtTimes",
    "MaxWeightsAtTimes",
    "TurnoverLimit",
    "MinCashBalance"
]


class BaseConstraint(CvxpyExpressionEstimator):
    """Base cvxpy constraint class."""


class BaseTradeConstraint(BaseConstraint):
    """Base class for constraints that operate on trades."""

    pass


class BaseWeightConstraint(BaseConstraint):
    """Base class for constraints that operate on weights.

    Here we can implement a method to pass benchmark weights
    and make the constraint relative to it rather than to the null
    portfolio.
    """

    pass

class MarketNeutral(BaseWeightConstraint):
    """Initial implementation of market neutrality.
    
    The benchmark portfolio weights are computed here 
    (weighting by rolling averages of the market volumes)
    but instead should be their own class (used as well
    by risk models, ...).
    """
    
    def __init__(self):
        self.covarianceforecaster = HistoricalFactorizedCovariance()
    
    def _pre_evaluation(self, universe, backtest_times):
        self.market_vector = cp.Parameter(len(universe)-1)
    
    def _values_in_time(self, t, past_volumes, past_returns, **kwargs):
        tmp = past_volumes.iloc[-250:].mean()
        tmp /= sum(tmp)
        
        tmp2 = self.covarianceforecaster.current_value @ (self.covarianceforecaster.current_value.T @ tmp)
        # print(tmp2)
        self.market_vector.value = np.array(tmp2)
        
    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        return w_plus[:-1].T @ self.market_vector == 0
        
        
class TurnoverLimit(BaseTradeConstraint):
    """Turnover limit as a fraction of the portfolio value.
    
    See page 37 of the book.
    
    :param delta: constant or changing in time turnover limit 
    :type delta: float or pd.Series 
    """
    
    def __init__(self, delta):
        self.delta = DataEstimator(delta, compile_parameter=True)
        
    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        return .5 * cp.norm1(z[:-1]) <= self.delta.parameter
        

class ParticipationRateLimit(BaseTradeConstraint):
    """A limit on maximum trades size as a fraction of market volumes.
    

    :param volumes: per-stock and per-day market volume estimates, or constant in time
    :type volumes: pd.Series or pd.DataFrame
    :param max_fraction_of_volumes: max fraction of market volumes that we're allowed to trade
    :type max_fraction_of_volumes: float, pd.Series, pd.DataFrame
    """

    def __init__(self, volumes, max_fraction_of_volumes=0.05):
        self.volumes = DataEstimator(volumes, compile_parameter=True)
        self.max_participation_rate = DataEstimator(
            max_fraction_of_volumes, compile_parameter=True)
        self.portfolio_value = cp.Parameter(nonneg=True)

    def _values_in_time(self, current_portfolio_value, **kwargs):
        self.portfolio_value.value = current_portfolio_value
        
    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Return a Cvxpy constraint."""
        return cp.multiply(cp.abs(z[:-1]), self.portfolio_value) <= cp.multiply(
            self.volumes.parameter, self.max_participation_rate.parameter
        )


class LongOnly(BaseWeightConstraint):
    """A long only constraint.
    
    Imposes that at each point in time the post-trade
    weights are non-negative.
    """

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Return a Cvxpy constraint."""
        return w_plus[:-1] >= 0


class NoTrade(BaseTradeConstraint):
    """No-trade condition on name on periods(s)."""
    
    def __init__(self, asset, periods):
        self.asset = asset
        self.periods = periods
    
    def _pre_evaluation(self, universe, backtest_times):
        self.index = universe.get_loc(self.asset)
        self.low = cp.Parameter()
        self.high = cp.Parameter()
    
    def _values_in_time(self, t, **kwargs):
        if t in self.periods:
            self.low.value = 0.
            self.high.value = 0.
        else:
            self.low.value = -100.
            self.high.value = +100.
    
    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        return [z[self.index] >= self.low, 
            z[self.index] <= self.high] 
        


class LeverageLimit(BaseWeightConstraint):
    """A limit on leverage.
    
    Leverage is defined as the :math:`\ell_1` norm of non-cash
    post-trade weights. Here we require that it is smaller than
    a given value

    :param limit: constant or varying in time leverage limit
    :type limit: float or pd.Series
    """

    def __init__(self, limit):
        self.limit = DataEstimator(limit, compile_parameter=True)

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Return a Cvxpy constraint."""
        return cp.norm(w_plus[:-1], 1) <= self.limit.parameter


class MinCashBalance(BaseWeightConstraint):
    """Requires that the cash account is larger than c_min dollars.
    
    This uses logic to subtract cash used as margin for the short 
    positions that is not documented in the book but is
    equivalent to the book definition's for long-only stock positions.
    """
    
    def __init__(self, c_min):
        self.c_min = DataEstimator(c_min)
        self.rhs = cp.Parameter()
    
    def _values_in_time(self, current_portfolio_value, **kwargs):
        self.rhs.value = self.c_min.current_value/current_portfolio_value
    
    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Return a Cvxpy constraint."""
        # TODO clarify this
        realcash = (w_plus[-1] - 2 * cp.sum(cp.neg(w_plus[:-1])))
        return realcash >= self.rhs
        
    
class LongCash(MinCashBalance):
    """Requires that cash be non-negative."""
    
    def __init__(self):
        super().__init__(0.)


class DollarNeutral(BaseWeightConstraint):
    """Long-short dollar neutral strategy."""

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Return a Cvxpy constraint."""
        return w_plus[-1] == 1


class MaxWeights(BaseWeightConstraint):
    """A max limit on weights.

    Attributes:
      limit: A series or number giving the weights limit.
    """

    def __init__(self, limit):
        self.limit = DataEstimator(limit, compile_parameter=True)

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Return a Cvxpy constraint."""
        return w_plus[:-1] <= self.limit.parameter


class MinWeights(BaseWeightConstraint):
    """A min limit on weights.

    Attributes:
      limit: A series or number giving the weights limit.
    """

    def __init__(self, limit):
        self.limit = DataEstimator(limit, compile_parameter=True)

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Return a Cvxpy constraint."""
        return w_plus[:-1] >= self.limit.parameter

class MinMaxWeightsAtTimes(BaseWeightConstraint):

    def __init__(self, limit, times):
        self.base_limit = limit
        self.times = times
    
    def _pre_evaluation(self, universe, backtest_times):
        self.backtest_times = backtest_times
        self.limit = cp.Parameter()
        
    def _values_in_time(self, t, mpo_step, **kwargs):
        tidx = self.backtest_times.get_loc(t)
        nowtidx = tidx + mpo_step
        if (nowtidx < len(self.backtest_times)) and self.backtest_times[nowtidx] in self.times:
            self.limit.value = self.base_limit
        else:
            self.limit.value = 100 * self.sign


class MinWeightsAtTimes(MinMaxWeightsAtTimes):
    
    sign = -1.

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Return a Cvxpy constraint."""
        return w_plus[:-1] >= self.limit
        
class MaxWeightsAtTimes(MinMaxWeightsAtTimes):

    sign = 1.
    
    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Return a Cvxpy constraint."""
        return w_plus[:-1] <= self.limit
        

class FactorMaxLimit(BaseWeightConstraint):
    """A max limit on portfolio-wide factor (e.g. beta) exposure.

    Args:
        factor_exposure: An (n * r) matrix giving the factor exposure per asset
        per factor, where n represents # of assets and r represents # of factors
        limit: A series of list or a single list giving the factor limits
    """

    def __init__(self, factor_exposure, limit):
        self.factor_exposure = DataEstimator(factor_exposure, compile_parameter=True)
        self.limit = DataEstimator(limit, compile_parameter=True)

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Return a Cvxpy constraint."""
        return self.factor_exposure.parameter.T @ w_plus[:-1] <= self.limit.parameter


class FactorMinLimit(BaseWeightConstraint):
    """A min limit on portfolio-wide factor (e.g. beta) exposure.

    Args:
        factor_exposure: An (n * r) matrix giving the factor exposure per asset
        per factor, where n represents # of assets and r represents # of factors
        limit: A series of list or a single list giving the factor limits
    """

    def __init__(self, factor_exposure, limit):
        self.factor_exposure = DataEstimator(factor_exposure, compile_parameter=True)
        self.limit = DataEstimator(limit, compile_parameter=True)

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Return a Cvxpy constraint."""
        return self.factor_exposure.parameter.T @ w_plus[:-1] >= self.limit.parameter


class FixedFactorLoading(BaseWeightConstraint):
    """A constraint to fix portfolio loadings to a set of factors.

    This can be used to impose market neutrality, a certain portfolio-wide alpha, ....

    Attributes:
        factor_exposure: An (n * r) matrix giving the factor exposure on each
        factor
        target: A series or number giving the targeted factor loading
    """

    def __init__(self, factor_exposure, target):
        self.factor_exposure = DataEstimator(factor_exposure, compile_parameter=True)
        self.target = DataEstimator(target, compile_parameter=True)

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Return a Cvxpy constraint."""
        return self.factor_exposure.parameter.T @ w_plus[:-1] == self.target.parameter



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains trading policies that can be backtested."""

import datetime as dt
import copy
import logging
import warnings

import pandas as pd
import numpy as np
import cvxpy as cp

from .costs import BaseCost
from .returns import BaseReturnsModel
from .constraints import BaseConstraint
from .estimator import PolicyEstimator, DataEstimator
from .errors import MissingValuesError, PortfolioOptimizationError
from .returns import ReturnsForecast, CashReturn
from .benchmark import *
from .utils import *

__all__ = [
    "Hold",
    "SellAll",
    "FixedTrades",
    "PeriodicRebalance",
    "ProportionalRebalance",
    "AdaptiveRebalance",
    "SinglePeriodOptimization",
    "MultiPeriodOptimization",
    "ProportionalTradeToTargets",
    "RankAndLongShort",
    "FixedWeights",
    "Uniform",
]


class BaseTradingPolicy(PolicyEstimator):
    """Base class for a trading policy."""


class Hold(BaseTradingPolicy):
    """Hold initial portfolio, don't trade."""

    def _values_in_time(self, current_weights, **kwargs):
        """Update sub-estimators and produce current estimate."""
        return pd.Series(0., index=current_weights.index)


class RankAndLongShort(BaseTradingPolicy):
    """Rank assets by signal; long highest and short lowest.

    Args:
        signal (pd.DataFrame): time-indexed DataFrame of signal for all symbols
            excluding cash. At each point in time the num_long assets with
            highest signal will have equal positive weight, and the num_short assets with
            lower signal will have equal negative weight. If two or more assets have the same
            signal value and they are on the boundary of either the top or bottom set,
            alphanumerical ranking will prevail.
        num_long (int or pd.Series): number of assets to long, default 1; if specified as Series
            it must be indexed by time.
        num_short (int or pd.Series): number of assets to short, default 1; if specified as Series
            it must be indexed by time.
        target_leverage (float or pd.Series): leverage of the resulting portfolio, default 1;
            if specified as Series it must be indexed by time.
    """

    def __init__(self, signal, num_long=1, num_short=1, target_leverage=1.):
        """Define sub-estimators at class attribute level."""
        self.num_long = DataEstimator(num_long)
        self.num_short = DataEstimator(num_short)
        self.signal = DataEstimator(signal)
        self.target_leverage = DataEstimator(target_leverage)

    def _values_in_time(self, t, current_weights, **kwargs):
        """Update sub-estimators and produce current estimate."""

        sorted_ret = pd.Series(
            self.signal.current_value, current_weights.index[:-1]
        ).sort_values()
        short_positions = sorted_ret.index[: self.num_short.current_value]
        long_positions = sorted_ret.index[-self.num_long.current_value:]

        target_weights = pd.Series(0., index=current_weights.index)
        target_weights[short_positions] = -1.
        target_weights[long_positions] = 1.

        target_weights /= sum(abs(target_weights))
        target_weights *= self.target_leverage.current_value

        # cash is always 1.
        target_weights[current_weights.index[-1]] = 1.

        return target_weights - current_weights


class ProportionalTradeToTargets(BaseTradingPolicy):
    """Given a DataFrame of target weights in time, trade in equal proportions to reach those.

    Initially, it loads the list of trading days and so at each day it knows
    how many are missing before the next target's day, and trades in equal proportions
    to reach those targets. If there are no targets remaining it defaults to
    not trading.

    Args:
        targets (pd.DataFrame): time-indexed DataFrame of target weight vectors at
            given points in time (e.g., start of each month).

    """

    def __init__(self, targets):
        self.targets = targets

    def _pre_evaluation(self, universe, backtest_times):
        """Get list of trading days."""
        self.trading_days = backtest_times

    def _values_in_time(self, t, current_weights, **kwargs):
        """Get current trade weights."""
        next_targets = self.targets.loc[self.targets.index >= t]
        assert np.allclose(next_targets.sum(1), 1.)
        if not len(next_targets):
            return pd.Series(0., index=current_weights.index)
        next_target = next_targets.iloc[0]
        next_target_day = next_targets.index[0]
        trading_days_to_target = len(self.trading_days[(
            self.trading_days >= t) & (self.trading_days < next_target_day)])
        return (next_target - current_weights) / (trading_days_to_target + 1)


class SellAll(BaseTradingPolicy):
    """Sell all assets to cash.

    This is useful to check the tcost model in the simulator.
    """

    def _values_in_time(self, t, current_weights, **kwargs):
        """Get current trade weights."""
        target = np.zeros(len(current_weights))
        target[-1] = 1.
        return target - current_weights


class FixedTrades(BaseTradingPolicy):
    """Each day trade the provided trade weights vector.

    If there are no weights defined for the given day, default to no
    trades.

    Args:
        trades_weights (pd.Series or pd.DataFrame): Series of weights
            (if constant in time) or DataFrame of trade weights
            indexed by time. It trades each day the corresponding vector.
    """

    def __init__(self, trades_weights):
        """Trade the tradevec vector (dollars) or tradeweight weights."""
        self.trades_weights = DataEstimator(trades_weights)

    def _recursive_values_in_time(self, t, current_weights, **kwargs):
        """We need to override recursion b/c we catch exception."""
        try:
            super()._recursive_values_in_time(t=t, current_weights=current_weights, **kwargs)
            return pd.Series(
                self.trades_weights.current_value,
                current_weights.index)
        except MissingValuesError:
            return pd.Series(0., current_weights.index)


class FixedWeights(BaseTradingPolicy):
    """Each day trade to the provided trade weights vector.

    If there are no weights defined for the given day, default to no
    trades.

    Args:
        target_weights (pd.Series or pd.DataFrame): Series of weights
            (if constant in time) or DataFrame of trade weights
            indexed by time. It trades each day to the corresponding vector.
    """

    def __init__(self, target_weights):
        """Trade the tradevec vector (dollars) or tradeweight weights."""
        self.target_weights = DataEstimator(target_weights)

    def _recursive_values_in_time(self, t, current_weights, **kwargs):
        """We need to override recursion b/c we catch exception."""
        try:
            super()._recursive_values_in_time(t=t, current_weights=current_weights, **kwargs)
            return pd.Series(self.target_weights.current_value,
                    current_weights.index) - current_weights
        except MissingValuesError:
            return pd.Series(0., current_weights.index)
            
            
class Uniform(FixedWeights):
    """Uniform allocation on non-cash assets."""
    
    def __init__(self):
        pass
    
    def _pre_evaluation(self, universe, backtest_times):
        target_weights = pd.Series(1., universe)
        target_weights.iloc[-1] = 0
        target_weights /= sum(target_weights)
        self.target_weights = DataEstimator(target_weights)


class PeriodicRebalance(FixedWeights):
    """Track a target weight vector rebalancing at given times.

    This calls `FixedWeights`. If you want to change the target in time
    use that policy directly.


    Args:
        target (pd.Series): portfolio weights to rebalance to.
        rebalancing_times (pd.DateTimeIndex): after the open trading on these days
            portfolio is equal to target.
    """

    def __init__(self, target, rebalancing_times):
        target_weights = pd.DataFrame(
            {el: target for el in rebalancing_times}).T
        super().__init__(target_weights)


class ProportionalRebalance(ProportionalTradeToTargets):
    """Track a target weight exactly at given times, trading proportionally to it each period.

    This calls `ProportionalTradeToTargets`. If you want to change the target in time
    use that policy directly.

    Args:
        target (pd.Series): portfolio weights to rebalance to.
        target_matching_times (pd.DateTimeIndex): after the open trading on these days
            portfolio is equal to target.
    """

    def __init__(self, target, target_matching_times):
        targets = pd.DataFrame({el: target for el in target_matching_times}).T
        super().__init__(targets)


class AdaptiveRebalance(BaseTradingPolicy):
    """Rebalance portfolio when deviates too far from target.

    We use the 2-norm as trigger for rebalance. You may want to
    calibrate ``tracking_error`` for your application
    by backtesting this policy, *e.g.*, to get your desired turnover.

    :param target: target weights to rebalance to.
        It is assumed a constant if it is a Series. If it varies in
        time (you must specify it for every trading day) pass a
        DataFrame indexed by time.
    :type target: pd.Series or pd.DataFrame
    :param tracking_error: we trade to match the target
        weights whenever the 2-norm of our weights minus the
        target is larger than this. Pass a Series if you want to vary it in
        time.
    :type tracking_error: pd.Series or pd.DataFrame
    
    """

    def __init__(self, target, tracking_error):
        self.target = DataEstimator(target)
        self.tracking_error = DataEstimator(tracking_error)

    def _values_in_time(self, t, current_weights, **kwargs):
        if np.linalg.norm(current_weights - self.target.current_value) > \
          self.tracking_error.current_value:
            return self.target.current_value - current_weights
        else:
            return pd.Series(0., current_weights.index)




class MultiPeriodOptimization(BaseTradingPolicy):
    r"""Multi Period Optimization policy.

    Implements the model developed in Chapter 5, in particular
    at page 49, of the book. You specify the objective terms
    using classes such as ReturnsForecast and TcostModel, each
    multiplied by its multiplier. You also specify lists
    of constraints. There are two ways to do it. You either
    define the same objective terms and costraints for each 
    step of the multi-period problem, or you define a different
    objective term and different list of constraints for each step.
    In addition we offer a `terminal_constraint` argument to
    simply impose that at the last step in the optimization the
    post-trade weights match the given weights (see page 51).
    
    When it computes the trajectory of weights for the future
    it only returns the first step (to the Simulator, typically).
    The future steps (planning horizon) are by default not returned.

    :param objective: these will be maximized;
        if you pass a single expression of BaseCost it is understood as the 
        same for all steps; if it's a list you must also pass a list of lists
        for `constraints`, each term represents the cost for each step of the optimization
        (starting from the first, i.e., today) and the length of the list is 
        used as planning_horizon (the value you pass there will be ignored) 
    :type objective: algebra of BaseCost or list of
    :param constraints: these will be
        imposed on the optimization. Default []. Pass this as a list of
        lists of the same length as `objective` to specify different 
        constraints at different time steps.
    :type constraints: list of BaseConstraints or list of those
    :param planning_horizon:  how many steps in the future we 
        plan for. Ignored if passing `objective` and `constraints` as lists.
        Default is None.
    :type planning_horizon: int or None
    :param terminal_constraint: if you pass a Series to this
        (default is None) it will impose that at the last step of the multi
        period optimization the post-trade weights are equal to this.
    :type terminal_constraint: pd.Series or None
    :param include_cash_return: whether to automatically include the ``CashReturn`` term in the objective,
        with default parameters. Default is ``True``.
    :type include_cash_return: bool
    :param benchmark: benchmark weights to use in the risk model and other terms that need it. Implemented
        ones are ``CashBenchmark``, the default, ``UniformBenchmark`` (uniform allocation on non-cash assets),
        and ``MarketBenchmark``, which approximates the market-weighted portfolio.
    :type benchmark: BaseBenchmark class or instance
    :param \**kwargs: these will be passed to cvxpy.Problem.solve,
        so you can choose your own solver and pass
        parameters to it.
    """

    def __init__(self, objective, constraints=[], include_cash_return=True, planning_horizon=None, terminal_constraint=None, benchmark=CashBenchmark, **kwargs):
        if hasattr(objective, '__iter__'):
            if not (hasattr(constraints, '__iter__') and len(constraints) and (hasattr(constraints[0], '__iter__') and len(objective) == len(constraints))):
                raise SyntaxError('If you pass objective as a list, constraints should be a list of lists of the same length.')
            self.planning_horizon = len(objective)
            self.objective = objective
            self.constraints = constraints
        else:
            if not np.isscalar(planning_horizon):
                raise SyntaxError('If `objective` and `constraints` are the same for all steps you must specify `planning_horizon`.')
            self.planning_horizon = planning_horizon
            self.objective = [copy.deepcopy(objective) for i in range(planning_horizon)] if planning_horizon > 1 else [objective]
            self.constraints = [copy.deepcopy(constraints) for i in range(planning_horizon)] if planning_horizon > 1 else [constraints]
        
        self.include_cash_return = include_cash_return
        if self.include_cash_return:
            self.objective = [el + CashReturn() for el in self.objective]
        self.terminal_constraint = terminal_constraint
        self.benchmark = benchmark() if isinstance(benchmark, type) else benchmark
        self.cvxpy_kwargs = kwargs

    def _compile_to_cvxpy(self):#, w_plus, z, value):
        """Compile all cvxpy expressions and the problem."""
        self.cvxpy_objective = [
            el._compile_to_cvxpy(self.w_plus_at_lags[i], self.z_at_lags[i], self.w_plus_minus_w_bm_at_lags[i]) 
            for i, el in enumerate(self.objective)]
        self.cvxpy_objective = sum(self.cvxpy_objective)
        assert self.cvxpy_objective.is_dcp()  # dpp=True)
        assert self.cvxpy_objective.is_concave()
        self.cvxpy_constraints = [
            flatten_heterogeneous_list([constr._compile_to_cvxpy(self.w_plus_at_lags[i], self.z_at_lags[i], self.w_plus_minus_w_bm_at_lags[i]) 
                for constr in el])
            for i, el in enumerate(self.constraints)]
        self.cvxpy_constraints = sum(self.cvxpy_constraints, [])
        self.cvxpy_constraints += [cp.sum(z) == 0 for z in self.z_at_lags]
        w = self.w_current
        for i in range(self.planning_horizon):
            self.cvxpy_constraints.append(self.w_plus_at_lags[i] == self.z_at_lags[i] + w)
            self.cvxpy_constraints.append(self.w_plus_at_lags[i] - self.w_bm == self.w_plus_minus_w_bm_at_lags[i])
            w = self.w_plus_at_lags[i]
        if not self.terminal_constraint is None:
            self.cvxpy_constraints.append(w == self.terminal_constraint)
        self.problem = cp.Problem(cp.Maximize(self.cvxpy_objective), self.cvxpy_constraints)
        assert self.problem.is_dcp()  # dpp=True)

    def _recursive_pre_evaluation(self, universe, backtest_times):
        """No point in using recursive super() method."""
        
        for obj in self.objective:
            obj._recursive_pre_evaluation(universe=universe, backtest_times=backtest_times)
        for constr_at_lag in self.constraints:
            for constr in constr_at_lag:
                constr._recursive_pre_evaluation(universe=universe, backtest_times=backtest_times)
        
        self.benchmark._recursive_pre_evaluation(universe=universe, backtest_times=backtest_times)
        self.w_bm = cp.Parameter(len(universe))

        # temporary
        # self.w_bm = np.zeros(len(universe))
        # self.w_bm[-1] = 1.
        
        # initialize the problem
        # self.portfolio_value = cp.Parameter(nonneg=True)
        self.w_current = cp.Parameter(len(universe))
        self.z_at_lags = [cp.Variable(len(universe)) for i in range(self.planning_horizon)] 
        self.w_plus_at_lags = [cp.Variable(len(universe)) for i in range(self.planning_horizon)]
        self.w_plus_minus_w_bm_at_lags = [cp.Variable(len(universe)) for i in range(self.planning_horizon)]

        # simulator will overwrite this with cached loaded from disk
        self.cache = {}

        # self._compile_to_cvxpy()#self.w_plus, self.z, self.portfolio_value)
        

    def _recursive_values_in_time(self, t, current_weights, current_portfolio_value, past_returns, past_volumes, current_prices, **kwargs):
        """Update all cvxpy parameters and solve."""
        
        assert current_portfolio_value > 0
        assert np.isclose(sum(current_weights), 1)
                
        for i, obj in enumerate(self.objective):
            obj._recursive_values_in_time(t=t, current_weights=current_weights, 
                    current_portfolio_value=current_portfolio_value, 
                    past_returns=past_returns, past_volumes=past_volumes, 
                    current_prices=current_prices, mpo_step=i, cache=self.cache, **kwargs)     
                     
        for i, constr_at_lag in enumerate(self.constraints):
            for constr in constr_at_lag:
                constr._recursive_values_in_time(t=t, current_weights=current_weights, 
                    current_portfolio_value=current_portfolio_value, 
                    past_returns=past_returns, past_volumes=past_volumes, 
                    current_prices=current_prices, mpo_step=i, cache=self.cache, **kwargs)   
                    
        self.benchmark._recursive_values_in_time(t=t, current_weights=current_weights, 
                    current_portfolio_value=current_portfolio_value, 
                    past_returns=past_returns, past_volumes=past_volumes, 
                    current_prices=current_prices, mpo_step=i, cache=self.cache, **kwargs) 
                    
        self.w_bm.value = self.benchmark.current_value
        self.w_current.value = current_weights.values

        try:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", message='Solution may be inaccurate')
                self.problem.solve(**self.cvxpy_kwargs)
        except cp.SolverError:
            raise PortfolioOptimizationError(
                f"Numerical solver for policy {self.__class__.__name__} at time {t} failed;"
                "try changing it, relaxing some constraints, or dropping some costs.")
        if self.problem.status in ["unbounded", "unbounded_inaccurate"]:
            raise PortfolioOptimizationError(
                f"Policy {self.__class__.__name__} at time {t} resulted in an unbounded problem."
            )
        if self.problem.status in ["infeasible", 'infeasible_inaccurate']:
            raise PortfolioOptimizationError(
                f"Policy {self.__class__.__name__} at time {t} resulted in an infeasible problem."
            )

        return pd.Series(self.z_at_lags[0].value, current_weights.index)


    
class SinglePeriodOptimization(MultiPeriodOptimization):
    r"""Single Period Optimization policy.

    Implements the model developed in Chapter 4, in particular
    at page 43, of the book. You specify the objective term
    using classes such as ReturnsForecast and TcostModel, each
    multiplied by its multiplier. You also specify a list
    of constraints.

    :param objective: this algebraic combination of cvxportfolio cost objects will be maximized
    :type objective: CombinedCost
    :param constraints: these will be imposed on the optimization. Default [].
    :type constraints: list of BaseConstraints
    :param include_cash_return: whether to automatically include the ``CashReturn`` term in the objective,
        with default parameters. Default is ``True``.
    :type include_cash_return: bool
    :param benchmark: benchmark weights to use in the risk model and other terms that need it. Implemented
        ones are ``CashBenchmark``, the default, ``UniformBenchmark`` (uniform allocation on non-cash assets),
        and ``MarketBenchmark``, which approximates the market-weighted portfolio.
    :type benchmark: BaseBenchmark class or instance
    :param \**kwargs: these will be passed to cvxpy.Problem.solve, so you can choose your own solver and pass
        parameters to it.
    """

    def __init__(self, objective, constraints=[], include_cash_return=True, benchmark=CashBenchmark, **kwargs):
        super().__init__([objective], [constraints], include_cash_return=include_cash_return, 
            benchmark=benchmark, **kwargs)
        




# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains classes that provide forecasts such as historical means
and covariances and are used internally by cvxportfolio objects. In addition,
forecast classes have the ability to cache results online so that if multiple
classes need access to the estimated value (as is the case in MultiPeriodOptimization
policies) the expensive evaluation is only done once. 
"""

import logging
from dataclasses import dataclass

import numpy as np

from .estimator import PolicyEstimator


def online_cache(_values_in_time):
    """A simple online cache that decorates _values_in_time.
    
    The instance it is used on needs to be hashable (we currently
    use the hash of its __repr__ via dataclass).
    """
    
    def wrapped(self, t, cache=None, **kwargs):
        
        if cache is None: # temporary to not change tests
            cache = {}
            
        if not (self in cache):
            cache[self] = {}
        
        if t in cache[self]:
            logging.debug(f'{self}._values_in_time at time {t} is retrieved from cache.')
            result = cache[self][t]
        else:
            result = _values_in_time(self, t=t, cache=cache, **kwargs)  
            logging.debug(f'{self}._values_in_time at time {t} is stored in cache.')
            cache[self][t] = result
        return result
        
    return wrapped

class BaseForecast(PolicyEstimator):
    """Base class for forecasters."""
    
    # def _recursive_pre_evaluation(self, universe, backtest_times):
    #     self.universe = universe
    #     self.backtest_times = backtest_times
    #
    
    
    def _agnostic_update(self, t, past_returns):
        """Choose whether to make forecast from scratch or update last one."""
        if (self.last_time is None) or (self.last_time != past_returns.index[-1]):
            logging.debug(f'{self}._values_in_time at time {t} is computed from scratch.')
            self._initial_compute(t=t, past_returns=past_returns)
        else:
            logging.debug(f'{self}._values_in_time at time {t} is updated from previous value.')
            self._online_update(t=t, past_returns=past_returns)
            
    def _initial_compute(self, t, past_returns):
        """Make forecast from scratch."""
        raise NotImplementedError

    def _online_update(self, t, past_returns):
        """Update forecast from period before."""
        raise NotImplementedError

@dataclass(unsafe_hash=True)
class HistoricalMeanReturn(BaseForecast):
    r"""Historical mean returns.
    
    This ignores both the cash returns column and all missing values.
    """
         
    def __post_init__(self):
        self.last_time = None
        self.last_counts = None
        self.last_sum = None
        
    def _pre_evaluation(self, universe, backtest_times):
        self.__post_init__()
    
    def _values_in_time(self, t, past_returns, cache=None, **kwargs):
        self._agnostic_update(t=t, past_returns=past_returns)
        return (self.last_sum / self.last_counts).values        
        
    def _initial_compute(self, t, past_returns):
        """Make forecast from scratch."""
        self.last_counts = past_returns.iloc[:, :-1].count()
        self.last_sum = past_returns.iloc[:, :-1].sum()
        self.last_time = t
        
    def _online_update(self, t, past_returns):
        """Update forecast from period before."""
        self.last_counts += ~(past_returns.iloc[-1, :-1].isnull())
        self.last_sum += past_returns.iloc[-1, :-1].fillna(0.)
        self.last_time = t

        
class HistoricalMeanError(BaseForecast):
    r"""Historical standard deviations of the mean of non-cash returns.
    
    For a given time series of past returns :math:`r_{t-1}, r_{t-2}, \ldots, r_0` this is
    :math:`\sqrt{\text{Var}[r]/t}`. When there are missing values we ignore them,
    both to compute the variance and the count.
    """

    def __init__(self):
        self.varianceforecaster = HistoricalVariance(kelly=False)
    
    def _values_in_time(self, t, past_returns, **kwargs):
        return np.sqrt(self.varianceforecaster.current_value / self.varianceforecaster.last_counts.values)
        
        
@dataclass(unsafe_hash=True)
class HistoricalVariance(BaseForecast):
    r"""Historical variances of non-cash returns.
    
    :param kelly: if ``True`` compute :math:`\mathbf{E}[r^2]`, else
        :math:`\mathbf{E}[r^2] - {\mathbf{E}[r]}^2`. The second corresponds
        to the classic definition of variance, while the first is what is obtained
        by Taylor approximation of the Kelly gambling objective. (See page 28 of the book.)
    :type kelly: bool
    """
    
    kelly: bool = True

    def __post_init__(self):
        if not self.kelly:
            self.meanforecaster = HistoricalMeanReturn()
        self.last_time = None
        self.last_counts = None
        self.last_sum = None
        
    def _pre_evaluation(self, universe, backtest_times):
        self.__post_init__()
    
    def _values_in_time(self, t, past_returns, **kwargs):
        self._agnostic_update(t=t, past_returns=past_returns)
        result = (self.last_sum / self.last_counts).values
        if not self.kelly:
            result -= self.meanforecaster.current_value**2
        return result
        
    def _initial_compute(self, t, past_returns):
        self.last_counts = past_returns.iloc[:, :-1].count()
        self.last_sum = (past_returns.iloc[:, :-1]**2).sum()
        self.last_time = t
        
    def _online_update(self, t, past_returns): #, last_estimation, last_counts, last_time):
        self.last_counts += ~(past_returns.iloc[-1, :-1].isnull())
        self.last_sum += past_returns.iloc[-1, :-1].fillna(0.)**2
        self.last_time = t


@dataclass(unsafe_hash=True)
class HistoricalFactorizedCovariance(BaseForecast):
    r"""Historical covariance matrix, sqrt factorized.
    
    :param kelly: if ``True`` compute each :math:` \Sigma_{i,j} \simeq \mathbf{E}[r^{i} r^{j}]`, else
        :math:` \Sigma_{i,j} \simeq \mathbf{E}[r^{i} r^{j}] - \mathbf{E}[r^{i}]\mathbf{E}[r^{j}]` matching
        the behavior of ``pandas.DataFrame.cov(ddof=0)`` (with same logic to handle missing data). 
        The second case corresponds to the classic definition of covariance, while the first is what is obtained
        by Taylor approximation of the Kelly gambling objective. (See page 28 of the book.)
    :type kelly: bool
    """
    
    kelly: bool = True
    
    def __post_init__(self):
        self.last_time = None
    
    def _pre_evaluation(self, universe, backtest_times):
        self.__post_init__()
    
    @staticmethod
    def _get_count_matrix(past_returns):
        r"""We obtain the matrix of non-null joint counts:
        
        .. math::
        
            \text{Count}\left(r^{i}r^{j} \neq \texttt{nan}\right).
        """
        tmp = (~past_returns.iloc[:,:-1].isnull()) * 1.
        return tmp.T @ tmp
    
    @staticmethod
    def _get_initial_joint_mean(past_returns):
        r"""Compute precursor of :math:`\Sigma_{i,j} = \mathbf{E}[r^{i}]\mathbf{E}[r^{j}]`.
        """
        nonnull = (~past_returns.iloc[:,:-1].isnull()) * 1.
        tmp = nonnull.T @ past_returns.iloc[:,:-1].fillna(0.)
        return tmp # * tmp.T

    @staticmethod
    def _factorize(Sigma):
        """Factorize matrix and remove negative eigenvalues."""
        eigval, eigvec = np.linalg.eigh(Sigma)
        eigval = np.maximum(eigval, 0.)
        return eigvec @ np.diag(np.sqrt(eigval))
        
    def _initial_compute(self, t, past_returns):
        self.last_counts_matrix = self._get_count_matrix(past_returns).values
        filled = past_returns.iloc[:,:-1].fillna(0.).values
        self.last_sum_matrix = filled.T @ filled
        if not self.kelly:
            self.joint_mean = self._get_initial_joint_mean(past_returns)

        self.last_time = t
        
    def _online_update(self, t, past_returns):
        nonnull = ~(past_returns.iloc[-1, :-1].isnull())
        self.last_counts_matrix += np.outer(nonnull, nonnull)
        last_ret = past_returns.iloc[-1, :-1].fillna(0.)
        self.last_sum_matrix += np.outer(last_ret, last_ret)
        self.last_time = t
        if not self.kelly:
            self.joint_mean += last_ret

    @online_cache
    def _values_in_time(self, t, past_returns, **kwargs):
  
        self._agnostic_update(t=t, past_returns=past_returns)
        Sigma = self.last_sum_matrix / self.last_counts_matrix
        
        if not self.kelly:
            tmp = self.joint_mean / self.last_counts_matrix
            Sigma -= tmp.T * tmp
               
        return self._factorize(Sigma) 



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module implements cost functions used by optimization-based policies.

Currently these are two: :class:`StocksTransactionCost` and :class:`StocksHoldingCost`.

The default parameters are chosen to approximate real costs for the stock market
as well as possible. 
"""

import cvxpy as cp
import numpy as np
import pandas as pd
import copy
import inspect

from .estimator import CvxpyExpressionEstimator,  DataEstimator
from .utils import periods_per_year
__all__ = ["HoldingCost", "TransactionCost", "StocksTransactionCost", "StocksHoldingCost"]


class BaseCost(CvxpyExpressionEstimator):
    """Base class for cost objects (and also risks).

    Here there is some logic used to implement the algebraic operations.
    See also :class:`CombinedCost`.
    """
    
    def _simulate(self, *args, **kwargs):
        """Simulate cost, used by market simulator.
        
        Look at its invocation in ``MarketSimulator`` for its list of 
        arguments.
        
        Cost classes that are meant to be used in the simulator
        should implement this.
        """
        raise NotImplementedError

    def __mul__(self, other):
        """Multiply by constant."""
        if not np.isscalar(other):
            raise SyntaxError("You can only multiply cost by a scalar.")
        return CombinedCosts([self], [other])

    def __add__(self, other):
        """Add cost expression to another cost expression.

        Idea is to create a new CombinedCost class that
        implements `_compile_to_cvxpy` and _recursive_values_in_time
        by summing over costs.

        """
        if isinstance(other, CombinedCosts):
            return other + self
        return CombinedCosts([self, other], [1.0, 1.0])

    def __rmul__(self, other):
        """Multiply by constant."""
        return self * other

    def __radd__(self, other):
        """Add cost expression to another cost."""
        return self + other

    def __neg__(self):
        """Take negative of cost."""
        return self * -1

    def __sub__(self, other):
        """Subtract other cost."""
        return self.__add__(-other)

    def __rsub__(self, other):
        """Subtract from other cost."""
        return other.__add__(-self)


class CombinedCosts(BaseCost):
    """Algebraic combination of :class:`BaseCost` instances.

    :param costs: instances of :class:`BaseCost`
    :type costs: list 
    :param multipliers: floats that multiply the ``costs``
    :type multipliers: list
    """

    def __init__(self, costs, multipliers):
        for cost in costs:
            if not isinstance(cost, BaseCost):
                raise SyntaxError("You can only sum `BaseCost` instances to other `BaseCost` instances.")
        self.costs = costs
        self.multipliers = multipliers
        
    def __add__(self, other):
        """Add other (combined) cost to self."""
        if isinstance(other, CombinedCosts):
            return CombinedCosts(self.costs + other.costs, self.multipliers + other.multipliers)
        else:
            return CombinedCosts(self.costs + [other], self.multipliers + [1.0])

    def __mul__(self, other):
        """Multiply by constant."""
        return CombinedCosts(self.costs, [el * other for el in self.multipliers])

    def _recursive_pre_evaluation(self, *args, **kwargs):
        """Iterate over constituent costs."""
        [el._recursive_pre_evaluation(*args, **kwargs) for el in self.costs]

    def _recursive_values_in_time(self, **kwargs):
        """Iterate over constituent costs."""
        [el._recursive_values_in_time(**kwargs) for el in self.costs ]

    def _compile_to_cvxpy(self, w_plus, z, portfolio_value):
        """Iterate over constituent costs."""
        self.expression = 0
        for multiplier, cost in zip(self.multipliers, self.costs):
            self.expression += multiplier * cost._compile_to_cvxpy(w_plus, z, portfolio_value) 
        return self.expression


class HoldingCost(BaseCost):
    """This is a generic holding cost model.
    
    Currently it is not meant to be used directly. Look at
    :class:`StocksHoldingCost` for its version specialized to
    the stock market.
    """

    def __init__(self, 
        spread_on_borrowing_assets_percent=None,
        spread_on_lending_cash_percent=None,
        spread_on_borrowing_cash_percent=None,
        periods_per_year=None,
        cash_return_on_borrow=False, #TODO revisit this plus spread_on_borrowing_stocks_percent syntax 
        dividends=None):
        
        self.spread_on_borrowing_assets_percent = None if spread_on_borrowing_assets_percent is None else \
            DataEstimator(spread_on_borrowing_assets_percent)
        self.dividends = None if dividends is None else DataEstimator(dividends, compile_parameter=True)        
        self.spread_on_lending_cash_percent = None if spread_on_lending_cash_percent is None else \
            DataEstimator(spread_on_lending_cash_percent)        
        self.spread_on_borrowing_cash_percent = None if spread_on_borrowing_cash_percent is None else \
            DataEstimator(spread_on_borrowing_cash_percent)

        self.periods_per_year = periods_per_year
        self.cash_return_on_borrow = cash_return_on_borrow
        
    def _pre_evaluation(self, universe, backtest_times):
        
        if self.spread_on_borrowing_assets_percent is not None or self.cash_return_on_borrow:
            self.borrow_cost_stocks = cp.Parameter(len(universe) - 1, nonneg=True)
        
        
    def _values_in_time(self, t, past_returns, **kwargs):
        """We use yesterday's value of the cash return here while in the simulator
        we use today's. In the US, updates to the FED rate are published outside
        of trading hours so we might as well use the actual value for today's. The difference
        is very small so for now we do this. 
        """
        ppy = periods_per_year(past_returns.index) if self.periods_per_year is None else \
            self.periods_per_year
                               
        cash_return = past_returns.iloc[-1,-1]

        if self.spread_on_borrowing_assets_percent is not None or self.cash_return_on_borrow:
            self.borrow_cost_stocks.value = np.ones(past_returns.shape[1] - 1) * (
                    cash_return if self.cash_return_on_borrow else 0.) + \
                self.spread_on_borrowing_assets_percent.current_value / (100 * ppy)
                
    def _simulate(self, t, h_plus, current_and_past_returns, **kwargs):
        
        ppy = periods_per_year(current_and_past_returns.index) if self.periods_per_year is None else \
            self.periods_per_year
        
        cash_return = current_and_past_returns.iloc[-1,-1]        
        multiplier = 1 / (100 * ppy)
        result = 0.
        borrowed_stock_positions = np.minimum(h_plus.iloc[:-1], 0.)
        result += np.sum(((cash_return if self.cash_return_on_borrow else 0.) + 
            (self.spread_on_borrowing_assets_percent._recursive_values_in_time(t) * multiplier if
            self.spread_on_borrowing_assets_percent is not None else 0.))
                 * borrowed_stock_positions)
        
        if self.dividends is not None:
            result += np.sum(h_plus[:-1] * self.dividends._recursive_values_in_time(t))
          
        # lending_spread = DataEstimator(spread_on_lending_cash_percent)._recursive_values_in_time(t) * multiplier
        # borrowing_spread = DataEstimator(spread_on_borrowing_cash_percent)._recursive_values_in_time(t) * multiplier

        # cash_return = current_and_past_returns.iloc[-1,-1]
        real_cash = h_plus.iloc[-1] + sum(np.minimum(h_plus.iloc[:-1], 0.))

        if real_cash > 0:
            if self.spread_on_lending_cash_percent is not None:
                result += real_cash * (max(cash_return - self.spread_on_lending_cash_percent._recursive_values_in_time(t) * multiplier, 0.) - cash_return)
        else:
            if self.spread_on_borrowing_cash_percent is not None:
                result += real_cash * self.spread_on_borrowing_cash_percent._recursive_values_in_time(t) * multiplier
            
        return result
            
            
    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile cost to cvxpy expression."""
        
        expression = 0. 
        
        if not (self.spread_on_borrowing_assets_percent is None):
           expression += cp.multiply(self.borrow_cost_stocks, cp.neg(w_plus)[:-1])
        
        if not (self.dividends is None):
            expression -= cp.multiply(self.dividends.parameter, w_plus[:-1])
        assert cp.sum(expression).is_convex()
        return cp.sum(expression)


class StocksHoldingCost(HoldingCost):    
    """A model for holding cost of stocks.

    :param spread_on_borrowing_stocks_percent: spread on top of cash return payed for borrowing assets,
        including cash. If ``None``, the default, it gets from :class:`Backtest` the
        value for the period.
    :type spread_on_borrowing_stocks_percent: float or pd.Series or pd.DataFrame or None
    :param spread_on_lending_cash_percent: spread that subtracts from the cash return for
        uninvested cash.
    :type spread_on_lending_cash_percent: float or pd.Series or None
    :param spread_on_borrowing_cash_percent: spread that adds to the cash return as rate payed for
        borrowed cash. This is not used as a policy optimization cost but is for the simulator cost.
    :type spread_on_borrowing_cash_percent: float or pd.Series or None
    :param dividends: dividends payed (expressed as fraction of the stock value)
        for each period.  If ``None``, the default, it gets from :class:`Backtest` the
        value for the period.
    :type dividends: pd.DataFrame or None
    :param periods_per_year: period per year (used in calculation of per-period cost). If None it is calculated
        automatically.
    :type periods_per_year: int or None
    :param cash_return_on_borrow: whether to add (negative of) cash return to borrow cost of assets
    :type cash_return_on_borrow: bool
    """
    
    
    def __init__(self, 
        spread_on_borrowing_stocks_percent=.5,
        spread_on_lending_cash_percent=.5,
        spread_on_borrowing_cash_percent=.5,
        periods_per_year=None,
        cash_return_on_borrow=True, #TODO revisit this plus spread_on_borrowing_stocks_percent syntax 
        dividends=0.):
        
        super().__init__(
            spread_on_borrowing_assets_percent=spread_on_borrowing_stocks_percent,
            spread_on_lending_cash_percent=spread_on_lending_cash_percent,
            spread_on_borrowing_cash_percent=spread_on_borrowing_cash_percent,
            periods_per_year=periods_per_year,
            cash_return_on_borrow=cash_return_on_borrow,
            dividends=dividends)

    
class TransactionCost(BaseCost):
    """This is a generic model for transaction cost of financial assets.
    
    Currently it is not meant to be used directly. Look at
    :class:`StocksTransactionCost` for its version specialized
    to the stock market.
    """

    def __init__(self, a=None, pershare_cost=None, b=0., window_sigma_est=250, window_volume_est=250, exponent=None):

        self.a = None if a is None else DataEstimator(a)
        self.pershare_cost = None if pershare_cost is None else DataEstimator(pershare_cost)
        self.b = None if b is None else DataEstimator(b)
        self.window_sigma_est = window_sigma_est
        self.window_volume_est = window_volume_est
        self.exponent = exponent
            
    def _pre_evaluation(self, universe, backtest_times):
        if self.a is not None or self.pershare_cost is not None:
            self.first_term_multiplier = cp.Parameter(len(universe)-1, nonneg=True)
        if self.b is not None:
            self.second_term_multiplier = cp.Parameter(len(universe)-1, nonneg=True)

    def _values_in_time(self, t,  current_portfolio_value, past_returns, past_volumes, current_prices, **kwargs):
        
        tmp = 0.
        
        if self.a is not None:
            _ = self.a.current_value 
            tmp += _ * np.ones(past_returns.shape[1]-1) if np.isscalar(_) else _ 
        if self.pershare_cost is not None:
            if current_prices is None:
                raise SyntaxError("If you don't provide prices you should set pershare_cost to None")
            tmp += self.pershare_cost.current_value / current_prices.values
        
        if self.a is not None or self.pershare_cost is not None:
            self.first_term_multiplier.value = tmp
        
        if self.b is not None:
            
            # TODO refactor this with forecast.py logic
            sigma_est = np.sqrt((past_returns.iloc[-self.window_sigma_est:, :-1]**2).mean()).values
            volume_est = past_volumes.iloc[-self.window_volume_est:].mean().values

            self.second_term_multiplier.value = self.b.current_value * sigma_est * \
                (current_portfolio_value / volume_est) ** ((2 if self.exponent is None else self.exponent) - 1)
                
    def _simulate(self, t, u, current_and_past_returns, current_and_past_volumes, current_prices, **kwargs):
        
        exponent = (1.5 if self.exponent is None else self.exponent)
        
        sigma = np.std(current_and_past_returns.iloc[-self.window_sigma_est:, :-1], axis=0)

        result = 0.
        if self.pershare_cost is not None:
            if current_prices is None:
                raise SyntaxError("If you don't provide prices you should set pershare_cost to None")
            result += self.pershare_cost._recursive_values_in_time(t) * int(sum(np.abs(u.iloc[:-1] + 1E-6) / current_prices.values))
        
        if self.a is not None:
            result += sum(self.a._recursive_values_in_time(t) * np.abs(u.iloc[:-1]))

        if self.b is not None:
            if current_and_past_volumes is None:
                raise SyntaxError("If you don't provide volumes you should set b to None")
            # we add 1 to the volumes to prevent 0 volumes error (trades are cancelled on 0 volumes)
            result += (np.abs(u.iloc[:-1])**exponent) @ (self.b._recursive_values_in_time(t)  *
                sigma / ((current_and_past_volumes.iloc[-1] + 1) ** (exponent - 1)))

        assert not np.isnan(result)
        assert not np.isinf(result)

        return -result
        
        
    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        
        expression = 0
        if self.a is not None or self.pershare_cost is not None:
            expression += cp.abs(z[:-1]).T @ self.first_term_multiplier
            assert expression.is_convex()
        if self.b is not None:
            expression += (cp.abs(z[:-1]) ** (
                    2 if self.exponent is None else self.exponent)
                ).T @ self.second_term_multiplier
            assert expression.is_convex()
        return expression
        
        
class StocksTransactionCost(TransactionCost):
    """A model for transaction costs of stocks.

    See pages 10-11 in `the book <https://stanford.edu/~boyd/papers/pdf/cvx_portfolio.pdf>`_.
    We don't include the short-term alpha term `c` here because it
    can be expressed with a separate `ReturnsForecast` object. 

    :param a: linear cost, which multiplies the absolute value of each trade. This can model
        (half) the bid-ask spread, or any fee linear in the size of a trade.
    :type a: float or pd.Series or pd.DataFrame
    :param pershare_cost: per-share trade cost, amount of dollars paid for each share traded. 
    :type pershare_cost: float or pd.Series or pd.DataFrame
    :param b: coefficient of the non-linear term of the transaction cost model, which multiplies
        the estimated volatility for each stock (see the book).  
    :type b: float or pd.Series or pd.DataFrame
    :param window_sigma_est: length of the window for the standard deviation of past returns used to 
        estimate the volatility :math:`\sigma` of each asset.
    :type window_sigma_est: int
    :param window_volume_est: length of the window for the mean of past volumes used as estimate
        of each period's volume. Has no effect on the simulator version of this which uses
        the actual volume.
    :type window_volume_est: int
    :param exponent: exponent of the non-linear term, defaults (if set to ``None``) to 1.5 for
        the simulator version, and 2 for the optimization version (because it is more efficient
        numerically and the difference is small, you can change it if you want).
    :type exponent: float or None
    """
    
    def __init__(self, a=0., pershare_cost=0.005, b=1.0, window_sigma_est=250, 
        window_volume_est=250, exponent=1.5):
        
        super().__init__(a=a, pershare_cost=pershare_cost, b=b, window_sigma_est=window_sigma_est,
            window_volume_est = window_volume_est, exponent=exponent)



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains classes that define return models for
portfolio optimization policies, and related objects.
"""

import cvxpy as cp
import numpy as np
import pandas as pd


from .costs import BaseCost, CombinedCosts
from .risks import BaseRiskModel
from .estimator import DataEstimator #, ParameterEstimator
from .forecast import HistoricalMeanReturn, HistoricalMeanError

__all__ = [
    "ReturnsForecast",
    "ReturnsForecastError",
    "CashReturn",
]


class BaseReturnsModel(BaseCost):
    """Base class for return models.

    Use this to define any logic common to return models.
    """


class CashReturn(BaseReturnsModel):
    r"""Objective term representing cash return.
    
    The forecast of cash return :math:`{\left(\hat{r}_t\right)}_n` is the observed value
    from last period :math:`{\left({r}_{t-1}\right)}_n`.
    
    This object is included automatically in :class:`SinglePeriodOptimization`
    and :class:`MultiPeriodOptimization` policies. You can change
    this behavior by setting their ``include_cash_return`` to False.
    
    :param short_margin_requirement: fraction of value of a short positions
        that is margined by portfolio cash
    :type short_margin_requirement: float
    """
    
    def __init__(self, cash_returns=None, short_margin_requirement=1.):
        self.cash_returns = None if cash_returns is None else DataEstimator(cash_returns, 
            compile_parameter=True, non_negative=True)
        self.short_margin_requirement = short_margin_requirement
        
    def _pre_evaluation(self, universe, backtest_times):
        self.cash_return_parameter = cp.Parameter(nonneg=True) if self.cash_returns is None \
            else self.cash_returns.parameter
            
        # else DataEstimator(self.cash_returns, non_negative=True, compile_parameter=True)
        
        
    def _values_in_time(self, t, past_returns, **kwargs):
        """Update cash return parameter as last cash return."""
        if self.cash_returns is None:
            self.cash_return_parameter.value = past_returns.iloc[-1,-1]
        
    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Apply cash return to "real" cash position (without shorts and margins)."""
        realcash = (w_plus[-1] - (1 + self.short_margin_requirement) * cp.sum(cp.neg(w_plus[:-1])))
        result = realcash * self.cash_return_parameter
        assert result.is_concave()
        return result
        
    
    
class ReturnsForecast(BaseReturnsModel):
    r"""Returns forecast for non-cash assets, provided by the user or computed from the data.

    This class represents the term :math:`\hat{r}_t`,
    the forecast of non-cash assets' returns at time :math:`t`.
    :ref:`Optimization-based policies` use this, typically as the first
    element of their objectives.
    See Chapters 4 and 5 of the `book <https://web.stanford.edu/~boyd/papers/pdf/cvx_portfolio.pdf>`_
    for more details.

    It can either get return forecasts from the user, for example
    computed with some machine learning technique, or it can estimate them
    automatically from the data.

    In the first case ``r_hat`` is specified as

    * :class:`float`, if :math:`\hat{r}_t` is the same for all times and non-cash assets
    * :class:`pandas.Series` with :class:`pandas.DatetimeIndex`, if :math:`\hat{r}_t` is the same for all assets but changes in time
    * :class:`pandas.Series` indexed by assets' names, if :math:`\hat{r}_t` is constant in time and changes across assets
    * :class:`pandas.DataFrame` with :class:`pandas.DatetimeIndex`, if :math:`\hat{r}_t` changes across time and assets.

    The returns' forecast provided by the user must be supplied for all assets
    excluding cash.

    If instead ``r_hat`` is not speficied it defaults to None. This instructs
    this class to compute :math:`\hat{r}_t` instead. It is
    done, at each step of a :class:`BackTest`, by evaluating the full average of the
    past returns (*i.e.,* the real returns :math:`{r}_{t-1}, {r}_{t-2}, \ldots`,
    where :math:`t` is the current time),  skipping :class:`numpy.nan` values.
    This is the default mode if no parameters are passed.

    :param r_hat: constant or time varying returns estimates, provided in the form of
        a pandas DataFrame indexed by timestamps of trading period and whose columns
        are all non-cash assets. Alternatively it can be a pandas Series indexed by the
        assets' names (so it is constant in time), a pandas Series indexed by time (so it is 
        constant across assets), or a float (constant for all times and assets). 
        If it is None, the default, the return forecasts are fitted from the data as historical means.
    :type r_hat: pd.Series or pd.DataFrame or float or None
    :param decay: decay factor used in :class:`MultiPeriodOptimization` policies.
        It is as a number in :math:`[0,1]`. At step :math:`\tau` of the MPO policy, where 
        :math:`\tau=t` is the initial one, the return predictions are multiplied by 
        :math:`\texttt{decay}^{\tau-t}`. So, ``decay`` close to zero models a `fast` signal
        while ``decay`` close to one a `slow` signal. The default value is 1.    
    :type decay: float

    :raises cvxportfolio.MissingValuesError: If the class accesses
        user-provided elements of ``r_hat`` that are :class:`numpy.nan`.

    :Example:

    >>> import cvxportfolio as cvx
    >>> policy = cvx.SinglePeriodOptimization(cvx.ReturnsForecast() - \
        0.5 * cvx.FullCovariance(), [cvx.LongOnly(), cvx.LeverageLimit(1)])
    >>> cvx.MarketSimulator(['AAPL', 'MSFT', 'GOOG']).backtest(policy).plot()

    Defines a single period optimization policy where the returns' forecasts
    :math:`\hat{r}_t` are the full average of past returns at each point in time
    and the risk model is the full covariance, also computed from the past returns.
    """

    def __init__(self, r_hat=None, decay=1.):
        
        if not r_hat is None:
            self.r_hat = DataEstimator(r_hat)
        else:
            self.r_hat = HistoricalMeanReturn()
        self.decay = decay
        
    def _pre_evaluation(self, universe, backtest_times):
        self.r_hat_parameter = cp.Parameter(len(universe)-1)
        
    def _values_in_time(self, t, past_returns, mpo_step=0, **kwargs):
        self.r_hat_parameter.value = self.r_hat.current_value * self.decay**(mpo_step)

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Cvxpy expression acts on non-cash assets."""
        return w_plus[:-1].T @ self.r_hat_parameter 
        


class ReturnsForecastError(BaseRiskModel):
    """Simple return forecast error risk with values provided by the user.

    Implements the model described in pages 31-32 of the paper. You
    must pass the delta Series (if constant) or DataFrame (if time-varying)
    of the forecast errors. Or, use one of the classes below to
    compute it automatically. Multiply this object by an external multiplier
    (which can itself be a ParameterEstimator) and calibrate for the right
    amount of penalization.

    :param deltas_errors: constant per-symbol
        errors on the returns forecasts (if Series),
        or varying in time (if DataFrame), 
        or fitted from the data as the standard deviation
        of the historical mean estimator 
    :type deltas_errors: pd.DataFrame or pd.Series or None
    """

    def __init__(self, deltas=None):
        
        if not deltas is None:
            self.deltas = DataEstimator(deltas)
        else:
            self.deltas = HistoricalMeanError()
            
    def _pre_evaluation(self, universe, backtest_times):
        self.deltas_parameter = cp.Parameter(len(universe)-1, nonneg=True)


    def _values_in_time(self, t, past_returns, **kwargs):
        self.deltas_parameter.value = self.deltas.current_value

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile to cvxpy expression."""
        return cp.abs(w_plus_minus_w_bm[:-1]).T @ self.deltas_parameter







# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for the cost objects."""

import unittest
from pathlib import Path

import cvxpy as cvx
import numpy as np
import pandas as pd

from cvxportfolio.costs import *
from cvxportfolio.returns import *
from cvxportfolio.risks import *
# from cvxportfolio.legacy import LegacyReturnsForecast #, MultipleReturnsForecasts



class TestCosts(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        """Load the data and initialize cvxpy vars."""
        # cls.sigma = pd.read_csv(Path(__file__).parent / "sigmas.csv", index_col=0, parse_dates=[0])
        cls.returns = pd.read_csv(Path(__file__).parent / "returns.csv", index_col=0, parse_dates=[0])
        cls.volumes = pd.read_csv(Path(__file__).parent / "volumes.csv", index_col=0, parse_dates=[0])
        cls.w_plus = cvx.Variable(cls.returns.shape[1])
        cls.w_plus_minus_w_bm = cvx.Variable(cls.returns.shape[1])
        cls.z = cvx.Variable(cls.returns.shape[1])
        cls.N = cls.returns.shape[1]
        
        
    def test_cost_algebra(self):
        # n = len(self.returns.columns)
        # wplus = cvx.Variable(n)
        self.w_plus_minus_w_bm.value = np.random.uniform(size=self.N)
        self.w_plus_minus_w_bm.value /= sum(self.w_plus_minus_w_bm.value)
        t = self.returns.index[1]

        cost1 = DiagonalCovariance()
        cost2 = FullCovariance(self.returns.iloc[:, :-1].T @ self.returns.iloc[:, :-1] / len(self.returns))
        cost3 = cost1 + cost2

        cost3._recursive_pre_evaluation(universe=self.returns.columns, backtest_times=self.returns.index)
        expr3 = cost3._compile_to_cvxpy(self.w_plus, self.z, self.w_plus_minus_w_bm)
        expr1 = cost1._compile_to_cvxpy(self.w_plus, self.z, self.w_plus_minus_w_bm)
        expr2 = cost2._compile_to_cvxpy(self.w_plus, self.z, self.w_plus_minus_w_bm)
        cost3._recursive_values_in_time(t=t, past_returns=self.returns.loc[self.returns.index < t])
        self.assertTrue(expr3.value == expr1.value + expr2.value)

        cost4 = cost1 * 2
        expr4 = cost4._compile_to_cvxpy(self.w_plus, self.z, self.w_plus_minus_w_bm)
        self.assertTrue(expr4.value == expr1.value * 2)

        cost3 = cost1 - cost2
        expr3 = cost3._compile_to_cvxpy(self.w_plus, self.z, self.w_plus_minus_w_bm)
        self.assertTrue(expr3.value == expr1.value - expr2.value)

        cost3 = -cost1 + 2 * cost2
        expr3 = cost3._compile_to_cvxpy(self.w_plus, self.z, self.w_plus_minus_w_bm)
        self.assertTrue( expr3.value == -expr1.value + 2 * expr2.value)

        cost3 = -cost1 + 2 * (cost2 + cost1)
        expr3 = cost3._compile_to_cvxpy(self.w_plus, self.z, self.w_plus_minus_w_bm)
        self.assertTrue( np.isclose(expr3.value, -expr1.value + 2 * (expr2.value + expr1.value)))

        cost3 = cost1 - 2 * (cost2 + cost1)
        expr3 = cost3._compile_to_cvxpy(self.w_plus, self.z, self.w_plus_minus_w_bm)
        self.assertTrue( expr3.value == expr1.value - 2 * (expr2.value + expr1.value))
        
    def test_hcost(self):
        """Test holding cost model."""
        dividends = pd.Series(np.random.randn(self.N-1), self.returns.columns[:-1])
        dividends *= 0.
        hcost = StocksHoldingCost(spread_on_borrowing_stocks_percent=.5,
                            dividends=dividends)
        
        t = 100 # this is picked so that periods_per_year evaluates to 252
        hcost._recursive_pre_evaluation(universe=self.returns.columns, backtest_times=self.returns.index)
        expression = hcost._compile_to_cvxpy(self.w_plus, self.z, self.w_plus_minus_w_bm)
        hcost._recursive_values_in_time(t=self.returns.index[t], past_returns=self.returns.iloc[:t])
        cash_ret = self.returns.iloc[t-1][-1]
        
        for i in range(10):
            self.w_plus.value = np.random.randn(self.N)
            self.w_plus.value[-1] = 1 - np.sum(self.w_plus.value[:-1])
        
            print(expression.value)

            print( -np.sum(np.minimum(self.w_plus.value[:-1], 0.)) * (cash_ret + .5/(100 * 252)))
            print( - self.w_plus.value[:-1].T @ dividends)
            print(-np.sum(np.minimum(self.w_plus.value[:-1], 0.)) * (cash_ret + 0.5/(100 * 252))
               - self.w_plus.value[:-1].T @ dividends)

            self.assertTrue(np.isclose(expression.value,
                -np.sum(np.minimum(self.w_plus.value[:-1], 0.)) * (cash_ret + 0.5/(100 * 252))
                #+ np.abs(self.w_plus.value[-1])* 0.5/(100 * 252)
               - self.w_plus.value[:-1].T @ dividends
            ))


    def test_tcost(self):
        """Test tcost model."""
        value = 1e6
        
        pershare_cost = pd.Series([0., 0.005, 0.], [self.returns.index[12], self.returns.index[23], self.returns.index[34]])
        b = pd.Series([0., 0., 1.], [self.returns.index[12], self.returns.index[23], self.returns.index[34]])
        
        tcost = StocksTransactionCost(
            a=0.001/2, pershare_cost=pershare_cost, b=b, window_sigma_est=250, window_volume_est=250, exponent=1.5)
        
        t = self.returns.index[12]
        
        tcost._recursive_pre_evaluation(universe=self.returns.columns, backtest_times=self.returns.index)
        expression = tcost._compile_to_cvxpy(self.w_plus, self.z, self.w_plus_minus_w_bm)
        
        # only spread
        
        tcost._recursive_values_in_time(t=self.returns.index[12], 
            current_portfolio_value = value,
            past_returns=self.returns.iloc[:12], 
            past_volumes=self.volumes.iloc[:12],
            current_prices=pd.Series(np.ones(self.returns.shape[1]-1), self.returns.columns[:-1]))
        
        
        self.z.value = np.random.randn(self.returns.shape[1])
        self.z.value[-1] = -np.sum(self.z.value[:-1])
        
        est_tcost_lin = sum(np.abs(self.z.value[:-1]) * 0.0005)
        print(est_tcost_lin)
        print(expression.value)
        self.assertTrue(np.isclose(expression.value, est_tcost_lin))
        
        
        # spread and fixed cost
        
        prices = pd.Series(np.random.uniform(1, 100, size=self.returns.shape[1]-1), self.returns.columns[:-1])
        
        tcost._recursive_values_in_time(t=self.returns.index[23], 
            current_portfolio_value = value,
            past_returns=self.returns.iloc[:23], 
            past_volumes=self.volumes.iloc[:23],
            current_prices=prices)
        
        
        self.z.value = np.random.randn(self.returns.shape[1])
        self.z.value[-1] = -np.sum(self.z.value[:-1])
        
        est_tcost_lin = sum(np.abs(self.z.value[:-1]) * 0.0005)
        est_tcost_lin += np.abs(self.z.value[:-1]) @ (0.005 / prices)
        print(est_tcost_lin)
        print(expression.value)
        self.assertTrue(np.isclose(expression.value, est_tcost_lin))
        
        # spread and nonlin cost
        
        tcost._recursive_values_in_time(t=self.returns.index[34], 
            current_portfolio_value = value,
            past_returns=self.returns.iloc[:34], 
            past_volumes=self.volumes.iloc[:34],
            current_prices=pd.Series(np.ones(self.returns.shape[1]-1), self.returns.columns[:-1]))
        
        
        self.z.value = np.random.randn(self.returns.shape[1])
        self.z.value[-1] = -np.sum(self.z.value[:-1])
        
        est_tcost_lin = sum(np.abs(self.z.value[:-1]) * 0.0005)
        volumes_est = self.volumes.iloc[:34].mean().values
        sigmas_est = np.sqrt((self.returns.iloc[:34, :-1]**2).mean()).values
        est_tcost_nonnlin = (np.abs(self.z.value[:-1])**(3/2)) @ (sigmas_est * np.sqrt(value / volumes_est))
        print(est_tcost_lin)
        print(est_tcost_nonnlin)
        print(expression.value)
        self.assertTrue(np.isclose(expression.value, est_tcost_lin+est_tcost_nonnlin))
        
        

if __name__ == '__main__':
    unittest.main()
        




# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for the return forecast objects."""

import unittest
from pathlib import Path

import cvxpy as cp
import numpy as np
import pandas as pd
import numpy as np



from cvxportfolio.returns import *

class TestReturns(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        """Load the data and initialize cvxpy vars."""
        # cls.sigma = pd.read_csv(Path(__file__).parent / "sigmas.csv", index_col=0, parse_dates=[0])
        cls.returns = pd.read_csv(Path(__file__).parent / "returns.csv", index_col=0, parse_dates=[0])
        # cls.volumes = pd.read_csv(Path(__file__).parent / "volumes.csv", index_col=0, parse_dates=[0])
        cls.w_plus = cp.Variable(cls.returns.shape[1])
        cls.w_plus_minus_w_bm = cp.Variable(cls.returns.shape[1])
        cls.z = cp.Variable(cls.returns.shape[1])
        cls.N = cls.returns.shape[1]
        
    def boilerplate(self, model):
        """Initialize objects, compile cvxpy expression."""
        model._recursive_pre_evaluation(universe=self.returns.columns, backtest_times=self.returns.index)
        return model._compile_to_cvxpy(self.w_plus, self.z, self.w_plus_minus_w_bm)
        
        
    def test_cash_returns(self):
        "Test CashReturn object with last return as forecast."
        cash_model = CashReturn()
        cvxpy_expression = self.boilerplate(cash_model)
        self.w_plus.value = np.random.randn(self.N)
        cash_model._recursive_values_in_time(t=None, past_returns=self.returns.iloc[:123])
        cr = self.returns.iloc[122, -1]
        self.assertTrue(cvxpy_expression.value == cr * (self.w_plus[-1].value + 2 * np.sum(np.minimum(self.w_plus[:-1].value, 0.))))
        
    def test_cash_returns_provided(self):
        "Test CashReturn object with provided cash returns."
        cash_model = CashReturn(self.returns.iloc[:,-1])
        cvxpy_expression = self.boilerplate(cash_model)
        self.w_plus.value = np.random.randn(self.N)
        cash_model._recursive_values_in_time(t=self.returns.index[123], past_returns=None)
        cr = self.returns.iloc[123, -1]
        self.assertTrue(cvxpy_expression.value == cr * (self.w_plus[-1].value + 2 * np.sum(np.minimum(self.w_plus[:-1].value, 0.))))
        
    def test_returns_forecast(self):
        "Test ReturnsForecast object with provided assets' returns."
        alpha_model = ReturnsForecast(self.returns.iloc[:,:-1])
        cvxpy_expression = self.boilerplate(alpha_model)
        alpha_model._recursive_values_in_time(t=self.returns.index[123], past_returns=None)
        self.w_plus.value = np.random.randn(self.N)
        print(cvxpy_expression.value)
        print(self.w_plus[:-1].value @ self.returns.iloc[123][:-1]) 
        #+ ((self.w_plus[-1].value + np.sum(np.minimum(self.w_plus[:-1].value, 0.))) * self.returns.iloc[123][-1]))
        self.assertTrue(np.isclose(cvxpy_expression.value, self.w_plus[:-1].value @ self.returns.iloc[123][:-1]))
                #+ ((self.w_plus[-1].value + 2 * np.sum(np.minimum(self.w_plus[:-1].value, 0.))) * self.returns.iloc[123][-1])))
        
        
    def test_full_returns_forecast(self):
        "Test ReturnsForecast object with historical mean forecasts."
        alpha_model = ReturnsForecast()
        cvxpy_expression = self.boilerplate(alpha_model)
        t = self.returns.index[123]
        alpha_model._recursive_values_in_time(t=t, past_returns = self.returns.loc[self.returns.index<t])
        self.w_plus.value = np.random.uniform(size=self.N)
        self.w_plus.value /= sum(self.w_plus.value)
        myforecast = self.returns.iloc[:, :-1].loc[self.returns.index < t].mean()
        # myforecast.iloc[-1] = self.returns.iloc[122, -1]
        self.assertTrue(np.isclose(cvxpy_expression.value, self.w_plus.value[:-1] @ myforecast))
        
    def test_returns_forecast_error(self):
        "Test ReturnsForecastError object with provided values."
        delta = self.returns.iloc[:, :-1].std(ddof=0) / np.sqrt(len(self.returns))
        # delta.iloc[-1] = 0
        error_risk = ReturnsForecastError(delta)
        cvxpy_expression = self.boilerplate(error_risk)
        error_risk._recursive_values_in_time(t='ciao', past_returns='hello')
        self.w_plus_minus_w_bm.value = np.random.randn(self.N)
        self.assertTrue(np.isclose(cvxpy_expression.value, np.abs(self.w_plus_minus_w_bm.value[:-1]) @ delta))


    def test_full_returns_forecast_error(self):
        "Test ReturnsForecastError object with as forecast the std of the mean estimator."
        error_risk = ReturnsForecastError()
        cvxpy_expression = self.boilerplate(error_risk)
        t = self.returns.index[123]
        past_returns = self.returns.loc[self.returns.index < t]
        error_risk._recursive_values_in_time(t=t, past_returns = past_returns)
        self.w_plus_minus_w_bm.value = np.random.randn(self.N)
        delta = past_returns.std(ddof=0) / np.sqrt(past_returns.count())
        print(cvxpy_expression.value)
        print(np.abs(self.w_plus_minus_w_bm.value[:-1]) @ delta[:-1])
        self.assertTrue(np.isclose(cvxpy_expression.value, np.abs(self.w_plus_minus_w_bm.value[:-1]) @ delta[:-1]))
    
    

if __name__ == '__main__':
    unittest.main()





# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module implements Estimator base classes. Policies, costs, and constraints inherit
from this.
"""

import numpy as np
import pandas as pd
from .errors import MissingValuesError, DataError
import cvxpy as cp


class Estimator:
    """Estimator base class.

    Policies, costs, and constraints inherit from this. When overloading
    methods defined here one should be careful on deciding whether to call
    the `super()` corresponding method. It can make sense to call it before
    some logic, after, or not calling it at all. Also, any subclass of this that uses
    logic defined here should be careful to put estimator subclasses at the class
    attribute level, so that the two methods defined here get called recursively
    on them.
    """

    def _recursive_values_in_time(self, **kwargs):
        """Evaluates estimator at a point in time recursively on its sub-estimators.

        This function is called by Simulator classes on Policy classes
        returning the current trades list. Policy classes, if
        they contain internal estimators, should declare them as attributes
        and call this base function (via `super()`) before
        they do their internal computation. CvxpyExpression estimators
        should instead define this method to update their Cvxpy parameters.
        
        Once we finalize the interface all parameters will be listed here.
        """
        for _, subestimator in self.__dict__.items():
            if hasattr(subestimator, "_recursive_values_in_time"):
                subestimator._recursive_values_in_time(**kwargs)
        if hasattr(self, "_values_in_time"):
            self.current_value = self._values_in_time(**kwargs)
            return self.current_value


class PolicyEstimator(Estimator):
    """Base class for (most) estimators that are part of policy objects."""
    

    def _recursive_pre_evaluation(self, universe, backtest_times):
        """Recursively initialize estimator tree for backtest.

        :param universe: names of assets to be traded 
        :type universe: pandas.Index
        :param backtest_times: times at which the estimator will be evaluated
        :type backtest_time: pandas.DatetimeIndex
        """
        for _, subestimator in self.__dict__.items():
            if hasattr(subestimator, "_recursive_pre_evaluation"):
                subestimator._recursive_pre_evaluation(universe, backtest_times)
        if hasattr(self, "_pre_evaluation"):
            self._pre_evaluation(universe, backtest_times)
    

class CvxpyExpressionEstimator(PolicyEstimator):
    """Base class for estimators that are Cvxpy expressions."""

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile term to cvxpy expression.

        This is called by a Policy class on its terms before the start of the backtest
        to compile its Cvxpy problem. If the Policy changes in time
        this is called at every time step.

        It can either return a scalar expression, in the case of objective terms,
        or a list of cvxpy constraints, in the case of constraints.

        In MultiPeriodOptimization policies this is called separately
        for costs and constraints at different look-ahead steps with
        the corresponding w_plus and z.
        
        
        :param w_plus: post-trade weights 
        :type w_plus: cvxpy.Variable
        :param z: trade weights 
        :type z: cvxpy.Variable
        :param w_plus_minus_w_bm: post-trade weights minus benchmark weights 
        :type w_plus_minus_w_bm: cvxpy.Variable
        """
        raise NotImplementedError


class DataEstimator(PolicyEstimator):
    """Estimator of point-in-time values from internal `self.data`.

    It also implements logic to check that no `np.nan` are returned
    by its `_recursive_values_in_time` method, which is the way `cvxportfolio`
    objects use this class to get data.

    Args:
        data (object, pandas.Series, pandas.DataFrame): Data expressed
            preferably as pandas Series or DataFrame where the first
            index is a pandas.DateTimeIndex. Otherwise you can
            pass a callable object which implements the _recursive_values_in_time method
            (with the standard signature) and returns the corresponding value in time,
             or a constant float, numpy.array, or even pandas Series or DataFrame not
            indexed by time (e.g., a covariance matrix where both index and columns
            are the stock symbols).
        use_last_available_time (bool): if the pandas index exists
            and is a pandas.DateTimeIndex you can instruct self._recursive_values_in_time
            to retrieve the last available value at time t by setting
            this to True. Default is False.

    """

    def __init__(self, data, use_last_available_time=False, allow_nans=False, 
        compile_parameter=False, non_negative=False, positive_semi_definite=False):
        self.data = data
        self.use_last_available_time = use_last_available_time
        self.allow_nans = allow_nans
        self.compile_parameter = compile_parameter
        self.non_negative = non_negative
        self.positive_semi_definite =positive_semi_definite
    
    def _recursive_pre_evaluation(self, universe, backtest_times):
        # super()._recursive_pre_evaluation(universe, backtest_times)
        if self.compile_parameter:
            value = self.internal__recursive_values_in_time(t=backtest_times[0])
            self.parameter = cp.Parameter(value.shape if hasattr(value, "shape") else (), 
                PSD=self.positive_semi_definite, nonneg=self.non_negative)
            

    def value_checker(self, result):
        """Ensure that only scalars or arrays without np.nan are returned.

        Args:
            result (int, float, or np.array): data produced by self._recursive_values_in_time

        Returns:
            result (int, float, or np.array): same data if no np.nan are present and type is correct

        Raises:
            cvxportfolio.MissingValuesError: if np.nan's are present in result
            cvxportfolio.DataError: if data is not in the right form
        """

        if np.isscalar(result):
            if np.isnan(result) and not self.allow_nans:
                raise MissingValuesError(
                    f"{self.__class__.__name__}._recursive_values_in_time result is a np.nan scalar."
                )
            else:
                return result

        if isinstance(result, np.ndarray):
            if np.any(np.isnan(result)) and not self.allow_nans:
                message = f"{self.__class__.__name__}._recursive_values_in_time result is an array with np.nan's."
                if hasattr(self.data, 'columns') and len(self.data.columns) == len(result):
                    message += "Specifically, the problem is with symbol(s): " + str(self.data.columns[np.isnan(result)])
                raise MissingValuesError(message)
            else:
                return result

        raise DataError(
            f"{self.__class__.__name__}._recursive_values_in_time result is not a scalar or array."
        )

    def internal__recursive_values_in_time(self, t, *args, **kwargs):
        """Internal method called by `self._recursive_values_in_time`."""

        if hasattr(self.data, "_recursive_values_in_time"):
            return self.value_checker(
                self.data._recursive_values_in_time(
                    t, *args, **kwargs))

        if (
            hasattr(self.data, "loc")
            and hasattr(self.data, "index")
            and (
                isinstance(self.data.index, pd.DatetimeIndex)
                or (
                    isinstance(self.data.index, pd.MultiIndex)
                    and isinstance(self.data.index.levels[0], pd.DatetimeIndex)
                )
            )
        ):
            try:
                if self.use_last_available_time:
                    if isinstance(self.data.index, pd.MultiIndex):
                        newt = self.data.index.levels[0][
                            self.data.index.levels[0] <= t
                        ][-1]
                    else:
                        newt = self.data.index[self.data.index <= t][-1]
                    tmp = self.data.loc[newt]
                else:
                    tmp = self.data.loc[t]
                if hasattr(tmp, "values"):
                    return self.value_checker(tmp.values)
                else:
                    return self.value_checker(tmp)

            except (KeyError, IndexError):
                raise MissingValuesError(
                    f"{self.__class__.__name__}._recursive_values_in_time could not find data for requested time."
                )

        if hasattr(self.data, "values"):
            return self.value_checker(self.data.values)

        return self.value_checker(self.data)

    def _recursive_values_in_time(self, t, *args, **kwargs):
        """Obtain value of `self.data` at time t or right before.

        Args:
            t (pandas.TimeStamp): time at which we evaluate the estimator

        Returns:
            result (float, numpy.array): if you use a callable object make
                sure that it returns a float or numpy array (and not,
                for example, a pandas object)

        """
        self.current_value = self.internal__recursive_values_in_time(t, *args, **kwargs)
        # we do this because in some cases they never get compiled
        if hasattr(self, 'parameter'): 
            self.parameter.value = self.current_value
        return self.current_value


# class ConstantEstimator(cvxpy.Constant, DataEstimator):
#     """Cvxpy constant that uses the pre_evalution method to be initialized."""
#
#     def _recursive_pre_evaluation(self, returns, volumes, start_time, end_time, **kwargs):
#         """You should call super().__init__ it here."""
#         raise NotImplementedError


# class KnownData(DataEstimator):
#     """Data known beforehand to use in backtest.
#
#     :param data: user-provided data (known beforehand) in the form of time-indexed
#         Series or DataFrame (points-in-time are used in the backtest),
#         non-time indexed Series of DataFrame (treated as constant in time),
#         or float. If time-indexed, it can be multi-indexed (for example for
#         covariance matrices at points in time).
#     :type data: pd.DataFrame, pd.Series, float
#     :param use_last_available_time: if True, use last available time in the data
#         at each point in the backtest
#     :type use_last_available_time: bool
#
#     :ivar value: current point-in-time value populated during a backtest
#
#     :raises MissingValueError: if data retrieved at a point in time contains nan's
#     """
#
#     def __init__(self, data, use_last_available_time=False):
#         self.data = data
#         self.use_last_available_time = use_last_available_time
#
#     def _recursive_values_in_time(self, t, **kwargs):
#         self.value = self.internal__recursive_values_in_time(t, *args, **kwargs)
#
#
# class KnownDataParameter(KnownData):
#     """Data known beforehand to use in backtest as Cvxpy parameter.
#
#     :param data: user-provided data (known beforehand) in the form of time-indexed
#         Series or DataFrame (points-in-time are used in the backtest),
#         non-time indexed Series of DataFrame (treated as constant in time),
#         or float. If time-indexed, it can be multi-indexed (for example for
#         covariance matrices at points in time).
#     :type data: pd.DataFrame, pd.Series, float
#     :param use_last_available_time: if True, use last available time in the data
#         at each point in the backtest
#     :type use_last_available_time: bool
#     :param cvxpy: if True, creates a Cvxpy parameter as self.parameter
#         with the shape of the provided data at the first point in time, and optional
#         attributes positive_semi_definite and non_negative
#     :type cvxpy: bool
#     :param positive_semi_definite: if True, make a PSD Cvxpy parameter
#     :type positive_semi_definite: bool
#     :param non_negative: if True, make a non negative Cvxpy parameter
#     :type non_negative: bool
#
#     :ivar parameter: Cvxpy parameter (if :param cvxpy: is True) with
#         value equal to :ivar value:
#
#     :raises MissingValueError: if data retrieved at a point in time contains nan's
#     """
#
#     def __init__(self, data, positive_semi_definite=False, non_negative=False, **kwargs):
#         super().__init__(data, **kwargs)
#         self.positive_semi_definite = positive_semi_definite
#         self.non_negative = non_negative
#
#     def _recursive_pre_evaluation(self, universe, backtest_times):
#         value = super().internal__recursive_values_in_time(t=backtest_times[0])
#         self.parameter = cp.Parameter(
#             value if hasattr(value, "shape") else (),
#             PSD=self.positive_semi_definite, nonneg=self.non_negative)
#
#     def _recursive_values_in_time(self, t, **kwargs):
#         self.parameter.value = self.internal__recursive_values_in_time(t, **kwargs)
        
        
#
# class ParameterEstimator(DataEstimator):
#     def __init__(self, *args, **kwargs):
#         super().__init__(self, *args, compile_parameter=True, **kwargs)
        
# class ParameterEstimator(cvxpy.Parameter, DataEstimator, PolicyEstimator):
#     """Data estimator of point-in-time values that contains a Cvxpy Parameter.
#
#     Attributes:
#         parameter (cvxpy.Parameter): the parameter object to use with cvxpy
#             expressions
#     Args:
#         same as cvxportfolio.DataEstimator
#
#     """
#
#     def __init__(self, data, positive_semi_definite=False, non_negative=False, use_last_available_time=False, allow_nans=False):
#         self.positive_semi_definite = positive_semi_definite
#         self.non_negative = non_negative
#         self.use_last_available_time = use_last_available_time
#         self.data = data
#         self.allow_nans = allow_nans
#         # super(DataEstimator).__init__(data, use_last_available_time)
#
#     def _recursive_pre_evaluation(self, universe, backtest_times):
#         """Use the start time of the simulation to initialize the Parameter."""
#         super()._recursive_pre_evaluation(universe, backtest_times)
#         value = super()._recursive_values_in_time(t=backtest_times[0])
#         super().__init__(value.shape if hasattr(value, "shape") else (),
#             PSD=self.positive_semi_definite, nonneg=self.non_negative)
#
#     def _recursive_values_in_time(self, t, **kwargs):
#         """Update Cvxpy Parameter value."""
#         self.value = super()._recursive_values_in_time(t=t, **kwargs)



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for the risk objects."""

import unittest
from pathlib import Path

import cvxpy as cp
import numpy as np
import pandas as pd

from cvxportfolio.risks import *

USED_RETURNS = 10

class TestRisks(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        """Load the data and initialize cvxpy vars."""
        cls.returns = pd.read_csv(Path(__file__).parent / "returns.csv", index_col=0, parse_dates=[0]).iloc[:, :USED_RETURNS]
        cls.w_plus = cp.Variable(cls.returns.shape[1])
        cls.w_plus_minus_w_bm = cp.Variable(cls.returns.shape[1])
        cls.z = cp.Variable(cls.returns.shape[1])
        cls.N = cls.returns.shape[1]
        
    def boilerplate(self, model):
        model._recursive_pre_evaluation(universe=self.returns.columns, backtest_times=self.returns.index)
        return model._compile_to_cvxpy(self.w_plus, self.z, self.w_plus_minus_w_bm)


    def test_full_sigma(self):
        historical_covariances = self.returns.iloc[:, :-1].rolling(50).cov(ddof=0).dropna()
        risk_model = FullCovariance(historical_covariances)
        
        cvxpy_expression = self.boilerplate(risk_model)

        self.assertTrue(cvxpy_expression.is_convex())

        t = historical_covariances.index[123][0]
        self.w_plus_minus_w_bm.value = np.random.randn(self.N)

        risk_model._recursive_values_in_time(t=t, past_returns='Hello!')

        self.assertTrue(np.isclose(cvxpy_expression.value, self.w_plus_minus_w_bm.value[:-1] @
                          historical_covariances.loc[t] @ self.w_plus_minus_w_bm.value[:-1]))
                               
    def test_full_estimated_sigma(self):

        risk_model = FullCovariance()
        
        cvxpy_expression = self.boilerplate(risk_model)
        self.assertTrue(cvxpy_expression.is_convex())

        # N = returns.shape[1]
        # returns.iloc[:, -1] = 0.

        # w_plus = cp.Variable(N)
        # risk_model._recursive_pre_evaluation(
        #     returns.iloc[:, :], None, start_time=returns.index[50], end_time=None)
        # cvxpy_expression = risk_model._compile_to_cvxpy(w_plus, None, None)
        
        t = pd.Timestamp('2014-06-02')
        past = self.returns.loc[self.returns.index < t]
        should_be = past.iloc[:,:-1].T @ past.iloc[:,:-1] / len(past)
        
        self.w_plus_minus_w_bm.value = np.random.randn(self.N) 
        
        risk_model._recursive_values_in_time(t=t, past_returns=past)
        print(cvxpy_expression.value)
        
        print(self.w_plus_minus_w_bm.value[:-1] @ should_be @ self.w_plus_minus_w_bm.value[:-1])
        
        self.assertTrue(np.isclose(cvxpy_expression.value,
                          self.w_plus_minus_w_bm.value[:-1] @ should_be @ self.w_plus_minus_w_bm.value[:-1]))
                          
    def test_diagonal_covariance(self):

        # N = returns.shape[1]
        # returns.iloc[:, -1] = 0.

        historical_variances = self.returns.iloc[:,:-1].rolling(50).var().shift(1).dropna()
        risk_model = DiagonalCovariance(historical_variances)
        cvxpy_expression = self.boilerplate(risk_model)
        self.assertTrue(cvxpy_expression.is_convex())

        t = historical_variances.index[123]
        self.w_plus_minus_w_bm.value = np.random.randn(self.N) 

        risk_model._recursive_values_in_time(t=t, past_returns='hello')

        self.assertTrue(np.isclose(cvxpy_expression.value, self.w_plus_minus_w_bm.value[:-1] @
                          np.diag(historical_variances.loc[t]) @ self.w_plus_minus_w_bm.value[:-1]))

    def test_full_diagonal_covariance(self):

        risk_model = DiagonalCovariance()
        cvxpy_expression = self.boilerplate(risk_model)
        self.assertTrue(cvxpy_expression.is_convex())

        t = pd.Timestamp('2014-06-02')
        past = self.returns.loc[self.returns.index < t]
        should_be = (past**2).mean()
        should_be.iloc[-1] = 0.
        
        risk_model._recursive_values_in_time(t=t, past_returns=past)
        self.w_plus_minus_w_bm.value = np.random.randn(self.N) 
        
        self.assertTrue(np.isclose(cvxpy_expression.value, self.w_plus_minus_w_bm.value @
                          np.diag(should_be) @ self.w_plus_minus_w_bm.value))


    def test_forecast_error(self):

        historical_variances = (self.returns.iloc[:, :-1]**2).rolling(50).mean().shift(1).dropna()
        
        risk_model = RiskForecastError(historical_variances)
        cvxpy_expression = self.boilerplate(risk_model)
        
        self.assertTrue(cvxpy_expression.is_convex())

        t = historical_variances.index[123]
        
        self.w_plus_minus_w_bm.value = np.random.randn(self.N) 

        risk_model._recursive_values_in_time(t=t, past_returns='hello')

        print(cvxpy_expression.value)
        print((np.abs(self.w_plus_minus_w_bm.value[:-1]) @ np.sqrt(historical_variances.loc[t]))**2)
        self.assertTrue(np.isclose(cvxpy_expression.value, 
        
        (np.abs(self.w_plus_minus_w_bm.value[:-1]) @ np.sqrt(historical_variances.loc[t]))**2
        
        ))

    def test_full_forecast_error(self):

        risk_model = RiskForecastError()
        cvxpy_expression = self.boilerplate(risk_model)
        self.assertTrue(cvxpy_expression.is_convex())

        t = pd.Timestamp('2014-06-02')
        
        past = self.returns.loc[self.returns.index < t]
        should_be = (past**2).mean()
        should_be.iloc[-1] = 0.

        risk_model._recursive_values_in_time(t=t, past_returns=past)
        self.w_plus_minus_w_bm.value = np.random.randn(self.N)

        self.assertTrue(np.isclose(cvxpy_expression.value, 
        
        (np.abs(self.w_plus_minus_w_bm.value) @ np.sqrt(should_be))**2
        
        ))
        
    def test_low_rank_covariance(self):
        
        F = pd.DataFrame(np.random.randn(2, self.N-1), columns=self.returns.columns[:-1])
        d = pd.Series(np.random.uniform(size=(self.N-1)), self.returns.columns[:-1])
        risk_model = FactorModelCovariance(F=F, d=d)
        
        cvxpy_expression = self.boilerplate(risk_model)
        self.assertTrue(cvxpy_expression.is_convex())
        
        risk_model._recursive_values_in_time(t=self.returns.index[12], past_returns='hello')
        self.w_plus_minus_w_bm.value = np.random.randn(self.N)
        
        self.assertTrue(np.isclose(cvxpy_expression.value, 
            self.w_plus_minus_w_bm.value[:-1] @ np.diag(d) @ self.w_plus_minus_w_bm.value[:-1] + \
                ((F @ self.w_plus_minus_w_bm.value[:-1])**2).sum()))
                
    def test_estimated_low_rank_covariance(self):
        
        risk_model = FactorModelCovariance()#normalize=False)
        
        cvxpy_expression = self.boilerplate(risk_model)
        self.assertTrue(cvxpy_expression.is_convex())
        
        t = pd.Timestamp('2014-06-02')
        
        past = self.returns.loc[self.returns.index < t]
        
        FS = past.T @ past / len(past)
        FS.iloc[:, -1] = 0.
        FS.iloc[-1, :] = 0.
        eigval, eigvec = np.linalg.eigh(FS)
        F = np.sqrt(eigval[-1]) * eigvec[:, -1]
        d = np.diag(FS) - F**2
        
        risk_model._recursive_values_in_time(t=t, past_returns=past)
        self.w_plus_minus_w_bm.value = np.random.randn(self.N)
        
        self.assertTrue(np.isclose(cvxpy_expression.value,
            self.w_plus_minus_w_bm.value @ np.diag(d) @ self.w_plus_minus_w_bm.value + \
                ((F @ self.w_plus_minus_w_bm.value)**2).sum()))
                
                
                
    def test_worst_case_risk(self):

        risk_model0 = FullCovariance()
        risk_model1 = DiagonalCovariance()
        worst_case = WorstCaseRisk([risk_model0, risk_model1])
        
        cvxpy_expression = self.boilerplate(worst_case)

        assert cvxpy_expression.is_convex()

        cvxpy_expression0 = risk_model0._compile_to_cvxpy(self.w_plus, self.z, self.w_plus_minus_w_bm)
        cvxpy_expression1 = risk_model1._compile_to_cvxpy(self.w_plus, self.z, self.w_plus_minus_w_bm)

        self.w_plus_minus_w_bm.value = np.ones(self.N)

        t = pd.Timestamp('2014-06-02')
        
        worst_case._recursive_values_in_time(t=t, past_returns=self.returns.loc[self.returns.index < t])

        print(cvxpy_expression.value)
        print(cvxpy_expression0.value)
        print(cvxpy_expression1.value)
        assert (cvxpy_expression.value == cvxpy_expression0.value)
        assert (cvxpy_expression.value > cvxpy_expression1.value)
        



if __name__ == '__main__':
    unittest.main()






# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for the data interfaces."""

import unittest
import shutil, tempfile
from pathlib import Path

import numpy as np
import pandas as pd

from cvxportfolio.data import (
    YfinanceBase,
    LocalDataStore,
    Yfinance,
    FredBase,
    FredRate,
    SqliteDataStore,
    TimeSeries,
    PickleStore,
)


class TestData(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        """Initialize data directory."""
        cls.datadir = Path(tempfile.mkdtemp())
        print('created', cls.datadir)
        
    @classmethod
    def tearDownClass(cls):
        """Remove data directory."""
        print('removing', cls.datadir)
        shutil.rmtree(cls.datadir)
        
        
    def test_time_series(self):
        ts = TimeSeries("ZM", base_location=self.datadir)
        assert not hasattr(ts, "data")
        ts._recursive_pre_evaluation()
        assert np.all(
            ts._recursive_values_in_time(pd.Timestamp("2023-04-11"), "foo", bar=None)
            == ts.data.loc["2023-04-11"]
        )
        
    def test_yfinance_download(self):
        """Test YfinanceBase."""

        data = YfinanceBase().download("AAPL", start="2023-04-01", end="2023-04-15")
        print(data)
        print(data.loc["2023-04-10"]["Return"])
        print(data.loc["2023-04-11", "Open"] / data.loc["2023-04-10", "Open"] - 1)
        self.assertTrue(np.isclose(
            data.loc["2023-04-10", "Return"],
            data.loc["2023-04-11", "Open"] / data.loc["2023-04-10", "Open"] - 1,
        ))
        self.assertTrue(np.isnan(data.iloc[-1]["Close"]))
    
    
    
    def test_fred(self):
        store = FredRate(self.datadir)
        data = store.update_and_load("DFF")
        print(data)
        self.assertTrue(np.isclose((1 + data["2023-04-10"]) **
                          store.trading_days, 1 + 4.83 / 100))
                          
                          
    def test_fred_base(self):
        data = FredBase().download("DFF")
        self.assertTrue(data["2023-04-06"] == 4.83)
        self.assertTrue(data.index[0] == pd.Timestamp("1954-07-01"))
        print(data)
        olddata = data.iloc[:-123]
        newdata = FredBase().download("DFF", olddata)
        self.assertTrue(np.all(data == newdata))
        

        
        
    def test_yfinance(self):
        """Test yfinance ability to store and retrieve."""

        store = Yfinance(self.datadir)
        data = store.update_and_load("ZM")

        print(data)

        self.assertTrue(np.isclose(
            data.loc["2023-04-05", "Return"],
            data.loc["2023-04-06", "Open"] / data.loc["2023-04-05", "Open"] - 1,
        ))

        data1 = store.update_and_load("ZM")
        print(data1)

        self.assertTrue(np.isnan(data1.iloc[-1]["Close"]))

        print((data1.iloc[: len(data) - 1].Return -
              data.iloc[:-1].Return).describe().T)

        self.assertTrue(np.allclose(
            data1.loc[data.index[:-1]].Return, data.iloc[:-1].Return))
    
    


    def test_sqlite3_store_series(self):
        """Test storing and retrieving of a Series with datetime index."""
        self.base_test_series(SqliteDataStore, self.datadir)

    def test_local_store_series(self):
        """Test storing and retrieving of a Series with datetime index."""
        self.base_test_series(LocalDataStore, self.datadir)
  
    def test_pickle_store_series(self):
        """Test storing and retrieving of a Series with datetime index."""
        self.base_test_series(PickleStore, self.datadir)
  
    
    def test_sqlite3_store_dataframe(self):
        """Test storing and retrieving of a DataFrame with datetime index."""
        self.base_test_dataframe(SqliteDataStore, self.datadir)


    def test_local_store_dataframe(self):
        """Test storing and retrieving of a DataFrame with datetime index."""
        self.base_test_dataframe(LocalDataStore, self.datadir)
    
    def test_pickle_store_dataframe(self):
        """Test storing and retrieving of a DataFrame with datetime index."""
        self.base_test_dataframe(PickleStore, self.datadir)
    

    def test_local_store_multiindex(self):
        """Test storing and retrieving of a DataFrame with datetime index."""
        self.base_test_multiindex(LocalDataStore, self.datadir)


    def test_sqlite3_store_multiindex(self):
        """Test storing and retrieving of a DataFrame with datetime index."""
        self.base_test_multiindex(SqliteDataStore, self.datadir)

    def test_pickle_store_multiindex(self):
        """Test storing and retrieving of a DataFrame with datetime index."""
        self.base_test_multiindex(PickleStore, self.datadir)  

    def base_test_series(self, storeclass, *args, **kwargs):
        """Test storing and retrieving of a Series with datetime index."""
        store = storeclass(*args, **kwargs)
        for data in [
            pd.Series(
                0.0,
                pd.date_range(
                    "2020-01-01",
                    "2020-01-10"),
                name="prova1"),
            pd.Series(3, pd.date_range("2020-01-01", "2020-01-10"), name="prova2"),
            pd.Series(
                "ciao", pd.date_range("2020-01-01", "2020-01-02", freq="H"), name="prova3"
            ),
            # test datetime conversion
            pd.Series(
                pd.date_range("2022-01-01", "2022-01-02", freq="H"),
                pd.date_range("2020-01-01", "2020-01-02", freq="H"),
                name="prova4",
            ),
            # test overwrite
            pd.Series(
                pd.date_range("2023-01-01", "2023-01-02", freq="H"),
                pd.date_range("2020-01-01", "2020-01-02", freq="H"),
                name="prova4",
            ),
        ]:
            print(data)
            print(data.index.dtype)
            print(data.dtypes)

            store.store(data.name, data)

            data1 = store.load(data.name)
            print(data1)
            print(data1.index.dtype)
            print(data1.dtypes)

            self.assertTrue(data.name == data1.name)
            self.assertTrue(all(data == data1))
            self.assertTrue(all(data.index == data1.index))
            self.assertTrue(data.dtypes == data1.dtypes)

        self.assertTrue(store.load("blahblah") is None)
        
        
    def base_test_dataframe(self, storeclass, *args, **kwargs):
        """Test storing and retrieving of a DataFrame with datetime index."""
        store = storeclass(*args, **kwargs)
        index = pd.date_range("2020-01-01", "2020-01-02", freq="H")
        data = {
            "one": range(len(index)),
            "two": np.arange(len(index)) / 19.0,
            "three": ["hello"] * len(index),
            "four": [np.nan] * len(index),
        }

        data["two"][2] = np.nan
        data = pd.DataFrame(data, index=index)
        print(data)
        print(data.index.dtype)
        print(data.dtypes)

        store.store("example", data)
        data1 = store.load("example")
        print(data1)
        print(data1.index.dtype)
        print(data1.dtypes)

        self.assertTrue(all(data == data1))
        self.assertTrue(all(data.index == data1.index))
        self.assertTrue(all(data.dtypes == data1.dtypes))
            
    def base_test_multiindex(self, storeclass, *args, **kwargs):
        """Test storing and retrieving of a Series or DataFrame with multi-index."""
        store = storeclass(*args, **kwargs)

        # second level is object
        timeindex = pd.date_range("2022-01-01", "2022-01-30")
        second_level = ["hello", "ciao", "hola"]
        index = pd.MultiIndex.from_product([timeindex, second_level])
        data = pd.DataFrame(np.random.randn(len(index), 10), index=index)
        data.columns = [
            "one",
            "two",
            "tre",
            "quattro",
            "cinque",
            "sei",
            "sette",
            "otto",
            "nove",
            "dieci",
        ]

        print(data.index)
        print(data)
        print(data.index.dtype)
        print(data.dtypes)

        store.store("example", data)
        data1 = store.load("example")

        print(data1.index)
        print(data1)
        print(data1.index.dtype)
        print(data1.dtypes)

        self.assertTrue(all(data == data1))
        self.assertTrue(all(data.index == data1.index))
        self.assertTrue(all(data.index.dtypes == data1.index.dtypes))
        self.assertTrue(all(data.dtypes == data1.dtypes))

        # second level is timestamp
        timeindex = pd.date_range("2022-01-01", "2022-01-30")
        second_level = pd.date_range("2022-01-01", "2022-01-03")
        index = pd.MultiIndex.from_product([timeindex, second_level])
        data = pd.DataFrame(np.random.randn(len(index), 10), index=index)
        data.columns = [
            "one",
            "two",
            "tre",
            "quattro",
            "cinque",
            "sei",
            "sette",
            "otto",
            "nove",
            "dieci",
        ]

        print(data.index)
        print(data)
        print(data.index.dtype)
        print(data.dtypes)

        store.store("example", data)
        data1 = store.load("example")

        print(data1.index)
        print(data1)
        print(data1.index.dtype)
        print(data1.dtypes)

        self.assertTrue(all(data == data1))
        self.assertTrue( all(data.index == data1.index))
        self.assertTrue( all(data.index.dtypes == data1.index.dtypes))
        self.assertTrue( all(data.dtypes == data1.dtypes))
        
if __name__ == '__main__':
    unittest.main()
        











                      
                      

















# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import numpy as np

__all__ = ['periods_per_year', 'resample_returns', 'flatten_heterogeneous_list']

def periods_per_year(idx):
    """Given a datetime pandas index return the periods per year."""
    return int(np.round(len(idx) / ((idx[-1] - idx[0]) / pd.Timedelta('365.24d'))))
    
def resample_returns(returns, periods):
    """Resample returns expressed over number of periods to single period."""
    return np.exp(np.log(1 + returns) / periods) - 1

def flatten_heterogeneous_list(l):
    """[1, 2, 3, [4, 5]] -> [1, 2, 3, 4, 5]"""
    return sum(([el] if not hasattr(el, '__iter__') 
        else el for el in l), [])


# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for the data and parameter estimator objects."""

import numpy as np
import pandas as pd
import unittest

from cvxportfolio.estimator import DataEstimator #, ParameterEstimator
from cvxportfolio.errors import MissingValuesError, DataError


class PlaceholderCallable:
    def __init__(self, value):
        self.value = value

    def _recursive_values_in_time(self, t, **kwargs):
        return self.value

class TestEstimators(unittest.TestCase):
        
        
    def test_callable(self):
        estimator = DataEstimator(PlaceholderCallable(1.0))
        time = pd.Timestamp("2022-01-01")
        self.assertEqual(estimator._recursive_values_in_time(time), 1.0)

        estimator = DataEstimator(PlaceholderCallable(np.nan))
        with self.assertRaises(MissingValuesError):
            estimator._recursive_values_in_time(t=time)

        data = np.arange(10.0)
        estimator = DataEstimator(PlaceholderCallable(data))
        self.assertTrue(np.all(estimator._recursive_values_in_time(t=time) == data))

        data[1] = np.nan
        with self.assertRaises(MissingValuesError):
            estimator._recursive_values_in_time(time)


    def test_scalar(self):
        time = pd.Timestamp("2022-01-01")

        estimator = DataEstimator(1.0)
        self.assertTrue(estimator._recursive_values_in_time(t=time) == 1.0)
        estimator = DataEstimator(1)
        self.assertTrue(estimator._recursive_values_in_time(t=time) == 1.0)

        estimator = DataEstimator(np.nan)
        with self.assertRaises(MissingValuesError):
            estimator._recursive_values_in_time(t=time)


    def test_array(self):
        time = pd.Timestamp("2022-01-01")
        data = np.arange(10.0)

        estimator = DataEstimator(data)
        self.assertTrue(np.all(estimator._recursive_values_in_time(t=time) == data))

        data[1] = np.nan
        estimator = DataEstimator(data)
        with self.assertRaises(MissingValuesError):
            estimator._recursive_values_in_time(t=time)


    def test_series_dataframe_notime(self):
        time = pd.Timestamp("2022-01-01")
        data = pd.Series(np.arange(10.0))
        estimator = DataEstimator(data)
        self.assertTrue(np.all(estimator._recursive_values_in_time(t=time) == data.values))

        data = pd.DataFrame(np.random.randn(3, 3))
        estimator = DataEstimator(data)
        self.assertTrue(np.all(estimator._recursive_values_in_time(t=time) == data.values))


    def test_series_timeindex(self):
        index = pd.date_range("2022-01-01", "2022-01-30")
        print(index)
        data = pd.Series(np.arange(len(index)), index)
        estimator = DataEstimator(data)

        print(estimator._recursive_values_in_time("2022-01-05"))
        self.assertTrue(estimator._recursive_values_in_time("2022-01-05") == data.loc["2022-01-05"])

        with self.assertRaises(MissingValuesError):
            estimator._recursive_values_in_time("2022-02-05")

        estimator = DataEstimator(data, use_last_available_time=True)
        self.assertTrue(estimator._recursive_values_in_time("2022-02-05") == data.iloc[-1])

        with self.assertRaises(MissingValuesError):
            estimator._recursive_values_in_time("2021-02-05")

        data["2022-01-05"] = np.nan
        estimator = DataEstimator(data)
        self.assertTrue(estimator._recursive_values_in_time("2022-01-04") == data.loc["2022-01-04"])
        with self.assertRaises(MissingValuesError):
            estimator._recursive_values_in_time("2022-01-05")


    def test_dataframe_timeindex(self):
        index = pd.date_range("2022-01-01", "2022-01-30")
        data = pd.DataFrame(np.random.randn(len(index), 10), index=index)
        estimator = DataEstimator(data)

        print(estimator._recursive_values_in_time("2022-01-05"))
        self.assertTrue(np.all(estimator._recursive_values_in_time(
            "2022-01-05") == data.loc["2022-01-05"]))

        with self.assertRaises(MissingValuesError):
            estimator._recursive_values_in_time("2021-01-05")

        estimator = DataEstimator(data, use_last_available_time=True)
        self.assertTrue(np.all(estimator._recursive_values_in_time("2022-02-05") == data.iloc[-1]))

        data.loc["2022-01-05", 3] = np.nan
        estimator = DataEstimator(data, use_last_available_time=True)
        with self.assertRaises(MissingValuesError):
            estimator._recursive_values_in_time("2021-01-05")


    def test_dataframe_multindex(self):
        timeindex = pd.date_range("2022-01-01", "2022-01-30")
        second_level = ["hello", "ciao", "hola"]
        index = pd.MultiIndex.from_product([timeindex, second_level])
        data = pd.DataFrame(np.random.randn(len(index), 10), index=index)
        print(data.index)
        estimator = DataEstimator(data)
        self.assertTrue( np.all(estimator._recursive_values_in_time(
            "2022-01-05") == data.loc["2022-01-05"]))

        estimator = DataEstimator(data, use_last_available_time=True)
        self.assertTrue( np.all(estimator._recursive_values_in_time(
            "2022-02-05") == data.loc["2022-01-30"]))
        self.assertTrue( np.all(estimator._recursive_values_in_time(
            "2022-01-05") == data.loc["2022-01-05"]))
        with self.assertRaises(MissingValuesError):
            estimator._recursive_values_in_time("2020-01-05")

        index = pd.MultiIndex.from_product([second_level, timeindex])
        data = pd.DataFrame(np.random.randn(len(index), 10), index=index)
        estimator = DataEstimator(data)
        assert np.all(estimator._recursive_values_in_time("2020-01-05") == data.values)


    def test_parameter_estimator(self):
        timeindex = pd.date_range("2022-01-01", "2022-01-30")
        second_level = ["hello", "ciao", "hola"]
        index = pd.MultiIndex.from_product([timeindex, second_level])
        data = pd.DataFrame(np.random.randn(len(index), 10), index=index)
        estimator = DataEstimator(data, compile_parameter=True)
        self.assertTrue( not hasattr(estimator, "parameter"))
        estimator._recursive_pre_evaluation(universe=None, backtest_times=timeindex)
        # assert hasattr(estimator, 'parameter')
        self.assertTrue( hasattr(estimator, "parameter"))
        estimator._recursive_values_in_time("2022-01-05")
        self.assertTrue( np.all(estimator.parameter.value == data.loc["2022-01-05"]))
        
if __name__ == '__main__':
    unittest.main()
    



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from pathlib import Path

import numpy as np
import pandas as pd
import cvxpy as cp
import cvxportfolio as cvx


class BaseTestClass(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        """Load the data and initialize cvxpy vars."""
        cls.sigma = pd.read_csv(Path(__file__).parent / "sigmas.csv", index_col=0, parse_dates=[0])
        cls.returns = pd.read_csv(Path(__file__).parent / "returns.csv", index_col=0, parse_dates=[0])
        cls.volumes = pd.read_csv(Path(__file__).parent / "volumes.csv", index_col=0, parse_dates=[0])
        cls.w_plus = cp.Variable(cls.returns.shape[1])
        cls.w_plus_minus_w_bm = cp.Variable(cls.returns.shape[1])
        cls.z = cp.Variable(cls.returns.shape[1])
        cls.N = cls.returns.shape[1]
    
    def boilerplate(self, model):
        model._recursive_pre_evaluation(universe=self.returns.columns, backtest_times=self.returns.index)
        return model._compile_to_cvxpy(self.w_plus, self.z, self.w_plus_minus_w_bm)


# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .estimator import DataEstimator 
import logging
import warnings

import scipy.linalg

import cvxpy as cp
import numpy as np
import pandas as pd

from .costs import BaseCost
from .forecast import HistoricalVariance, HistoricalFactorizedCovariance

logger = logging.getLogger(__name__)


__all__ = [
    "FullCovariance",
    "DiagonalCovariance",
    "FactorModelCovariance",
    "RiskForecastError",
    "WorstCaseRisk",
]




class BaseRiskModel(BaseCost):
    pass




class FullCovariance(BaseRiskModel):
    """Quadratic risk model with full covariance matrix.
    
    :param Sigma: DataFrame of covariance matrices
        supplied by the user, or None if fitting from the past data.
        The DataFrame can either represents a single constant covariance matrix
        or one for each point in time.
    :type Sigma: pandas.DataFrame or None
    
    
    """
    # r"""Quadratic risk model with full covariance matrix.
    #
    # This class represents the term :math:`\Sigma_t`, *i.e.,*
    # the :math:`(n-1) \times (n-1)` positive semi-definite matrix
    # which estimates the covariance of the (non-cash) assets' returns.
    # :ref:`Optimization-based policies` use this, as is explained
    # in Chapter 4 and 5 of the `book <https://web.stanford.edu/~boyd/papers/pdf/cvx_portfolio.pdf>`_.
    #
    # The user can either supply a :class:`pandas.DataFrame` with the covariance matrix
    # (constant or varying in time) computed externally (for example
    # with some machine learning technique) or let this class estimate the covariance from the data.
    # The latter is the default behavior.
    #
    # This class implements three ways to compute the covariance matrix from the past returns. The
    # computation is repeated at each point in time :math:`t` of a :class:`BackTest` using only
    # the past returns available at that point: :math:`r_{t-1}, r_{t-2}, \ldots`.
    #
    # * *rolling covariance*, using :class:`pandas.DataFrame.rolling.cov`. This is done
    #   if the user specifies the ``rolling`` argument.
    # * *exponential moving window covariance*, using :class:`pandas.DataFrame.ewm.cov`. This is done
    #   if the user specifies the ``halflife`` argument (``rolling`` takes precedence).
    # * *full historical covariance*, using :class:`pandas.DataFrame.cov`. This is the default
    #   behavior if no arguments are specified.
    #
    # If there are missing data in the historical returns the estimated covariance may not
    # be positive semi-definite. We correct it by projecting on the positive semi-definite
    # cone (*i.e.*, we set the negative eigenvalues of the resulting :math:`\Sigma_t` to zero).
    #
    # :param Sigma: :class:`pandas.DataFrame` of covariance matrices
    #     supplied by the user. The DataFrame either represents a single (constant) covariance matrix
    #     or one for each point in time. In the latter case the DataFrame must have a :class:`pandas.MultiIndex`
    #     where the first level is a :class:`pandas.DatetimeIndex`. If ``None`` (the default)
    #     the covariance matrix is computed from past returns.
    # :type Sigma: pandas.DataFrame or None
    # :param rolling: if it is not ``None`` the covariance matrix will be estimated
    #     on a rolling window of size ``rolling`` of the past returns.
    # :type rolling: int or None
    # :param halflife: if it is not ``None`` the covariance matrix will be estimated
    #     on an exponential moving window of the past returns with half-life ``halflife``.
    #     If ``rolling`` is specified it takes precedence over ``halflife``. If both are ``None`` the full history
    #     will be used for estimation.
    # :type halflife: int or None
    # :param kappa: the multiplier for the associated forecast error risk
    #     (see pages 32-33 of the `book <https://web.stanford.edu/~boyd/papers/pdf/cvx_portfolio.pdf>`_).
    #     If ``float`` a passed it is treated as a constant, if ``pandas.Series`` with ``pandas.DateTime`` index
    #     it varies in time, if ``None`` the forecast error risk term will not be compiled.
    # :type kappa: float or pandas.Series or None
    # :param kelly: correct the covariance matrix with the term :math:`\mu\mu^T`, as is explained
    #     in page 28 of the `book <https://web.stanford.edu/~boyd/papers/pdf/cvx_portfolio.pdf>`_,
    #     to match the second term of the Taylor expansion of the portfolio log-return. Default
    #     is ``False``, corresponding to classical mean-variance optimization. If ``True``, it
    #     estimates :math:`\mu` with the same technique as :math:`\Sigma`, *i.e.*, with rolling window
    #     average, exponential moving window average, or an average of the full history.
    # :type kelly: bool
    # """

    def __init__(self, Sigma=None, kelly=True):

        if not Sigma is None:
            self.Sigma = DataEstimator(Sigma)
            self.alreadyfactorized = False
        else:
            self.Sigma = HistoricalFactorizedCovariance(kelly=kelly) 
            self.alreadyfactorized = True
            
    def _pre_evaluation(self, universe, backtest_times):
        self.Sigma_sqrt = cp.Parameter((len(universe)-1, len(universe)-1))

    def _values_in_time(self, t, past_returns, **kwargs):
        """Update forecast error risk here, and take square root of Sigma."""
        
        if self.alreadyfactorized:
            self.Sigma_sqrt.value = self.Sigma.current_value
        else:
            Sigma = self.Sigma.current_value
            eigval, eigvec = np.linalg.eigh(Sigma)
            eigval = np.maximum(eigval, 0.)
            self.Sigma_sqrt.value = eigvec @ np.diag(np.sqrt(eigval))

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        self.cvxpy_expression = cp.sum_squares(self.Sigma_sqrt.T @ w_plus_minus_w_bm[:-1])
        return self.cvxpy_expression

class RiskForecastError(BaseRiskModel):
    """Risk forecast error. 
    
    Implements the model defined in page 31 of the book. Takes same arguments
    as :class:`DiagonalCovariance`.
    
    :param sigma_squares: per-stock variances, indexed by time if DataFrame.
        If None it will be fitted on past data.
    :type sigma_squares: pd.DataFrame or pd.Series or None
    """

    def __init__(self, sigma_squares=None):
        if sigma_squares is None:
            self.sigma_squares = HistoricalVariance(kelly=True) #None None
        else:
            self.sigma_squares = DataEstimator(sigma_squares)
        # self.standard_deviations = ParameterEstimator(standard_deviations)
        # self.zeroforcash=True
        # self.kelly=True
        
    def _pre_evaluation(self, universe, backtest_times):
        self.sigmas_parameter = cp.Parameter(len(universe)-1, nonneg=True)#+self.kelly))

    def _values_in_time(self, t, past_returns, **kwargs):
        """Update forecast error risk here, and take square root of Sigma."""
        
        # if self.sigma_squares is None:
        #     sigma_squares = past_returns.var(ddof=0)
        #     if self.kelly:
        #         mean = past_returns.mean()
        #         sigma_squares += mean**2
        #     if self.zeroforcash:
        #         sigma_squares.iloc[-1] = 0.
        #     sigma_squares = sigma_squares.values
        # else:
        #     sigma_squares = self.sigma_squares.current_value
            
        sigma_squares = self.sigma_squares.current_value
        
        self.sigmas_parameter.value = np.sqrt(sigma_squares)

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):

        return cp.square(cp.abs(w_plus_minus_w_bm[:-1]).T @ self.sigmas_parameter)
                

class DiagonalCovariance(BaseRiskModel):
    """Diagonal covariance matrix, user-provided or fit from data.

    :param sigma_squares: per-stock variances, indexed by time if DataFrame.
        If None it will be fitted on past data.
    :type sigma_squares: pd.DataFrame or pd.Series or None 
    """

    def __init__(self, sigma_squares=None):
        if not sigma_squares is None:
            self.sigma_squares = DataEstimator(sigma_squares)
        else:
            self.sigma_squares = HistoricalVariance(kelly=True) #None
        #self.zeroforcash = True
        #self.kelly = True
        # self.standard_deviations = ParameterEstimator(standard_deviations)
        
    def _pre_evaluation(self, universe, backtest_times):
        self.sigmas_parameter = cp.Parameter(len(universe)-1) #+self.kelly))

    def _values_in_time(self, t, past_returns, **kwargs):
        """Update forecast error risk here, and take square root of Sigma."""
        #super()._recursive_values_in_time(t, current_weights, current_portfolio_value, past_returns, past_volumes, **kwargs)
        
        # if self.sigma_squares is None:
        #     sigma_squares = past_returns.var(ddof=0)
        #     if self.kelly:
        #         mean = past_returns.mean()
        #         sigma_squares += mean**2
        #     if self.zeroforcash:
        #         sigma_squares[-1] = 0.
        #     sigma_squares = sigma_squares.values
        # else:
        #     sigma_squares = self.sigma_squares.current_value
        
        sigma_squares = self.sigma_squares.current_value

        self.sigmas_parameter.value = np.sqrt(sigma_squares)

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):

        return cp.sum_squares(cp.multiply(w_plus_minus_w_bm[:-1], self.sigmas_parameter))


class FactorModelCovariance(BaseRiskModel):
    """Factor model covariance, either user-provided or fitted from the data.
    
    It has the structure
    
    :math:`F F^T + \mathbf{diag}(d)`
    
    where :math:`F` is a *tall* matrix (many more rows than columns) and the vector
    :math:`d` is all non-negative. 
    
    :param F: exposure matrices either constant or varying in time; if so, use a pandas multiindexed
         dataframe. If None it will be fitted.
    :type F: pd.DataFrame or None
    :param d: idyosyncratic variances either constant or varying in time; If None it will be fitted.
    :type d: pd.Series or pd.DataFrame or None
    :param num_factors: number of factors (columns of F), used if fitting the model
    :type num_factors: int    
    """

    # Args:
    #     exposures (pd.DataFrame): constant factor exposure matrix or a dataframe
    #         where the first index is time.
    #     idyosync (pd.DataFrame or pd.Series): idyosyncratic variances for the symbol,
    #         either fixed (pd.Series) or through time (pd.DataFrame).
    #     factor_Sigma (pd.DataFrame or None): a constant factor covariance matrix
    #         or a DataFrame with multiindex where the first index is time. If None,
    #         the default, it is understood that the factor covariance is the identity.
    #         (Otherwise we compute its matrix square root at each step internally and
    #          apply it to the exposures).
    #     forecast_error_kappa (float or pd.Series): uncertainty on the
    #         assets' correlations. See the paper, pages 32-33.

    # """

    factor_Sigma = None

    def __init__(self, F=None, d=None, num_factors=1, kelly=True):#, normalize=False):
        self.F = F if F is None else DataEstimator(F, compile_parameter=True) 
        self.d = d if d is None else DataEstimator(d) 
        if (self.F is None) or (self.d is None):
            self.fit = True
            self.Sigma = HistoricalFactorizedCovariance(kelly=kelly) #Sigma
        else:
            self.fit = False
        self.num_factors = num_factors

    def _pre_evaluation(self, universe, backtest_times):
        self.idyosync_sqrt_parameter = cp.Parameter(len(universe)-1)
        effective_num_factors = min(self.num_factors, len(universe)-1)
        self.F_parameter = cp.Parameter((effective_num_factors, len(universe)-1)) if self.F is None else self.F.parameter


    def _values_in_time(self, t, past_returns, **kwargs):
        
        if self.fit:
            Sigmasqrt = self.Sigma.current_value
            # numpy eigendecomposition has largest eigenvalues last
            self.F_parameter.value = Sigmasqrt[:, -self.num_factors:].T
            d = np.sum(Sigmasqrt[:, :-self.num_factors]**2, axis=1)
        else:
            d = self.d.current_value
        self.idyosync_sqrt_parameter.value = np.sqrt(d)


    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        self.expression = cp.sum_squares(cp.multiply(self.idyosync_sqrt_parameter, w_plus_minus_w_bm[:-1]))
        assert self.expression.is_dcp(dpp=True)

        self.expression += cp.sum_squares(self.F_parameter @ w_plus_minus_w_bm[:-1])
        assert self.expression.is_dcp(dpp=True)

        return self.expression
        

class WorstCaseRisk(BaseRiskModel):
    """Select the most restrictive risk model for each value of the allocation vector.

    Given a list of risk models, penalize the portfolio allocation by the
    one with highest risk value at the solution point. If uncertain about
    which risk model to use this procedure can be an easy solution.

    :param riskmodels: risk model instances on which to compute the worst-case
        risk.
    :type riskmodels: list 
    """

    def __init__(self, riskmodels):
        self.riskmodels = riskmodels

    def _recursive_pre_evaluation(self, universe, backtest_times):
        """Initialize objects."""
        for risk in self.riskmodels:
            risk._recursive_pre_evaluation(universe, backtest_times)

    def _recursive_values_in_time(self, **kwargs):
        """Update parameters."""
        for risk in self.riskmodels:
            risk._recursive_values_in_time(**kwargs)

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        risks = [risk._compile_to_cvxpy(w_plus, z, w_plus_minus_w_bm)
                 for risk in self.riskmodels]
        return cp.max(cp.hstack(risks))



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module defines benchmark portfolio weights in time that are used, for example,
by risk terms of optimization-based policies.
"""

import numpy as np
import pandas as pd

from .estimator import PolicyEstimator

__all__ = ['CashBenchmark', 'UniformBenchmark', 'MarketBenchmark']


class BaseBenchmark(PolicyEstimator):
    """Base class for cvxportfolio benchmark weights."""
    pass

class CashBenchmark(BaseBenchmark):
    """Default benchmark weights for cvxportfolio risk models.
    """

    def _pre_evaluation(self, universe, backtest_times):
        """Define current_value as a constant."""
        self.current_value = np.zeros(len(universe))
        self.current_value[-1] = 1.


class UniformBenchmark(BaseBenchmark):
    """Benchmark weights uniform on non-cash assets.
    """

    def _pre_evaluation(self, universe, backtest_times):
        """Define current_value as a constant."""
        self.current_value = np.ones(len(universe))
        self.current_value[-1] = 0.
        self.current_value /= np.sum(self.current_value[:-1])
   
     
class MarketBenchmark(BaseBenchmark):
    """Portfolio weighted by last year's total volumes.
    """
    
    def _values_in_time(self, past_volumes, **kwargs):
        """Update current_value using past year's volumes."""
        sumvolumes = past_volumes.loc[past_volumes.index >= (past_volumes.index[-1] - pd.Timedelta('365d'))].mean()
        result = np.zeros(len(sumvolumes) + 1)
        result[:-1] = sumvolumes / sum(sumvolumes)
        return result
        